from __future__ import annotations

from typing import TYPE_CHECKING, Any

from anytree import LevelGroupOrderIter, NodeMixin, RenderTree, findall
from munch import Munch

if TYPE_CHECKING:
    from ekfsm.devices.generic import Device

    from .slots import Slot


class SysTree(NodeMixin):
    """
    Base class for all system components including Hardware Modules and Devices.
    """

    def __init__(self, name: str, abort: bool = False) -> None:
        from ekfsm.log import ekfsm_logger

        self.logger = ekfsm_logger(name)
        self.name = name
        self.abort = abort

    def _render_tree(self) -> str:
        output = ""
        for pre, _, node in RenderTree(self):
            output += f"{pre}{repr(node)}" + "\n"
        return output

    def __str__(self):
        return self._render_tree()

    def print(self) -> None:
        print(self)


class HWModule(SysTree):
    """
    A HWModule represents an instantiation of a specifc hw board type,
    for example an instance of an EKF SC9 board.
    """

    def __init__(
        self,
        instance_name: str,
        config: dict,
        slot: Slot,
        abort: bool = False,
        *args,
        **kwargs,
    ) -> None:
        from ekfsm.core.utils import deserialize_hardware_tree, deserialize_module

        from .slots import SlotType

        super().__init__(instance_name, abort=abort)
        self._slot: Slot = slot
        self.config = config

        self.id, self.board_type, slot_type = deserialize_module(self.logger, config)
        self.children = deserialize_hardware_tree(self.logger, self.config, parent=self)

        self.slot_type = SlotType.from_string(slot_type)

        for children in LevelGroupOrderIter(self):
            for child in children:
                for sub_child in child.children:
                    setattr(child, sub_child.name.lower(), sub_child)

                    # If the device provides functions, add function and corresponding attributes to this hw module
                    self._create_functions_from_node_providers(sub_child)

    @property
    def is_master(self) -> bool:
        slot_attributes = self.slot.attributes
        if slot_attributes is not None and hasattr(slot_attributes, "is_master"):
            return slot_attributes.is_master
        return False

    def _create_functions_from_node_providers(self, device: Device) -> None:
        providers: Munch | None = getattr(device, "provides", None)

        if providers is not None:
            for key, value in providers.items():
                if key not in self.__dict__:
                    provider = Munch()
                else:
                    provider = getattr(self, key)

                provider.update(value)

                setattr(self, key, provider)

    def probe(self, *args, **kwargs) -> bool:
        from ekfsm.core.probe import ProbeableDevice

        nodes = findall(self, lambda node: isinstance(node, ProbeableDevice))
        for node in nodes:
            try:
                if node.probe(*args, **kwargs):
                    return True
            except Exception as e:
                self.logger.error(f"Error probing {node}: {e}")

        return False

    def info(self) -> dict[str, Any]:
        """
        Returns a dictionary with information about the hardware module.
        """
        return {
            "name": self.instance_name,
            "slot": self.slot,
        }

    @property
    def instance_name(self) -> str:
        if self.name is None:
            raise RuntimeError("Instance name not set")
        return self.name

    @property
    def slot(self) -> Slot:
        if self._slot is None:
            raise RuntimeError("Slot not set")
        return self._slot

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.name}, type={self.board_type})"



import socket
import struct
from pathlib import Path
from typing import List
from unittest.mock import patch

from smbus2 import SMBus

from ekfsm.devices.gpio import EKFIdSimGpio, SimGpio
from ekfsm.devices.smbus import SimSMBus

from .core.components import SysTree
from .core.sysfs import set_sysfs_root
from .devices import GPIO

GPIO_SIM_MAPPING: dict[str, SimGpio] = {}
SMBUS_SIM_MAPPING: dict[str, SimSMBus] = {}


def register_gpio_sim(major: int, minor: int, sim_gpio: SimGpio) -> None:
    name = f"{major}:{minor}"
    if name in GPIO_SIM_MAPPING:
        raise ValueError(f"GPIO_SIM_MAPPING already contains {name}")
    GPIO_SIM_MAPPING[name] = sim_gpio


def find_gpio_dev_with_major_minor(major: int, minor: int) -> SimGpio:
    name = f"{major}:{minor}"
    if name not in GPIO_SIM_MAPPING:
        raise ValueError(f"GPIO_SIM_MAPPING does not contain {name}")
    return GPIO_SIM_MAPPING[name]


def register_smbus_sim(bus_num: int, i2c_addr: int, sim_smbus: SimSMBus) -> None:
    name = f"{bus_num}:{i2c_addr}"
    if name in SMBUS_SIM_MAPPING:
        raise ValueError(f"SMBUS_SIM_MAPPING already contains {name}")
    SMBUS_SIM_MAPPING[name] = sim_smbus


def find_smbus_dev(bus_num: int, i2c_addr: int) -> SimSMBus:
    name = f"{bus_num}:{i2c_addr}"

    if name not in SMBUS_SIM_MAPPING:
        raise ValueError(f"SMBUS_SIM_MAPPING does not contain {name}")
    return SMBUS_SIM_MAPPING[name]


class GPIOSimulator(GPIO):
    def __init__(
        self,
        name: str,
        parent: SysTree | None = None,
        *args,
        **kwargs,
    ):
        super(GPIO, self).__init__(
            name,
            parent,
            *args,
            **kwargs,
        )
        major, minor = self._find_gpio_dev(parent, *args, **kwargs)
        self._sim_gpio = find_gpio_dev_with_major_minor(major, minor)
        self.number = minor

    def num_lines(self) -> int:
        return self._sim_gpio.num_lines()

    def set_pin(self, pin: int, value: bool) -> None:
        self._sim_gpio.set_pin(pin, value)

    def get_pin(self, pin: int) -> bool:
        return self._sim_gpio.get_pin(pin)

    def set_direction(self, pin: int, direction: bool) -> None:
        self._sim_gpio.set_direction(pin, direction)

    def __str__(self) -> str:
        return f"GPIO_SIM({self.name})"


class SMBusSimulator:
    def __init__(self, bus_num: int):
        self.bus_num = bus_num

    def read_word_data(self, i2c_addr: int, cmd: int) -> int:
        return find_smbus_dev(self.bus_num, i2c_addr).read_word_data(cmd)

    def read_block_data(self, i2c_addr: int, cmd: int) -> List[int]:
        return find_smbus_dev(self.bus_num, i2c_addr).read_block_data(cmd)

    def write_block_data(self, i2c_addr: int, cmd: int, data: List[int]):
        find_smbus_dev(self.bus_num, i2c_addr).write_block_data(cmd, data)

    def write_byte(self, i2c_addr: int, cmd: int):
        find_smbus_dev(self.bus_num, i2c_addr).write_byte(cmd)

    def write_word_data(self, i2c_addr: int, cmd: int, data: int):
        find_smbus_dev(self.bus_num, i2c_addr).write_word_data(cmd, data)


def patch_methods(cls, simulator, methods):
    patched = []
    for i, method in enumerate(methods):
        if hasattr(cls, method) and hasattr(simulator, method):
            patched.append(patch.object(cls, method, new_callable=lambda: getattr(simulator, method)))
            patched[i].start()


def enable_simulation(sysfs_path: Path | str) -> None:
    global GPIO_SIM_MAPPING
    GPIO_SIM_MAPPING = {}

    global SMBUS_SIM_MAPPING
    SMBUS_SIM_MAPPING = {}

    if isinstance(sysfs_path, str):
        sysfs_path = Path(sysfs_path)

    set_sysfs_root(sysfs_path)
    patch_methods(
        GPIO,
        GPIOSimulator,
        ["__init__", "num_lines", "set_pin", "get_pin", "set_direction", "__str__"],
    )
    patch_methods(
        SMBus,
        SMBusSimulator,
        [
            "__init__",
            "read_word_data",
            "read_block_data",
            "write_block_data",
            "write_byte",
            "write_word_data",
        ],
    )


def register_gpio_simulations():
    register_gpio_sim(233, 1, EKFIdSimGpio(0x38, 0x1, 0x0, 0x6))  # SRF Rev 0
    register_gpio_sim(233, 2, EKFIdSimGpio(0x34, 0xA, 0x0, 0x1))  # CCU Rev 0


class SocketSmbus(SimSMBus):
    def __init__(self, host: str, port: int) -> None:
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.connect((host, port))

    def read_word_data(self, cmd: int) -> int:
        req = struct.pack("BB", 0x4, cmd)
        self.sock.send(req)
        data = self.sock.recv(2)
        return struct.unpack("<H", data)[0]

    def read_block_data(self, cmd: int) -> List[int]:
        req = struct.pack("BB", 0x1, cmd)
        self.sock.send(req)
        _count = self.sock.recv(1)
        count = struct.unpack("B", _count)[0]
        data = self.sock.recv(count)
        return [int(data[i]) for i in range(0, len(data), 1)]

    def write_block_data(self, cmd: int, data: List[int]):
        _data = bytes(data)
        hdr = struct.pack("BBB", 0x2, cmd, len(_data))
        self.sock.send(hdr + _data)

    def write_byte(self, cmd: int):
        hdr = struct.pack("BB", 0x3, cmd)
        self.sock.send(hdr)

    def write_word_data(self, cmd: int, data: int):
        hdr = struct.pack("BBH", 0x5, cmd, data)
        self.sock.send(hdr)



import logging
import pprint
from pathlib import Path

from ekfsm.simctrl import enable_simulation, register_gpio_simulations
from ekfsm.system import System

logging.basicConfig(level=logging.DEBUG, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")

enable_simulation(Path(__file__).parent / "sim" / "sys")
register_gpio_simulations()

config = Path(__file__).parent / "sim" / "test_system.yaml"
system = System(config, abort=True)
system.print()

pprint.pprint(f"System slots {system.slots}")

cpu = system[1]
cpu.print()

print(f"CPU inventory: {cpu.inventory.vendor()} {cpu.inventory.model()} {cpu.inventory.serial()}")

srf = system["FAN"]
srf.print()

psu = system.psu
psu.print()

psu = system.slots["PSU_SLOT"].hwmodule
psu.print()

print(f"Slot info {system.slots['PSU_SLOT'].info()}")


print(f"PSU main: {psu.main.voltage()}V  {psu.main.current()}A")
print(f"PSU sby: {psu.sby.voltage()}V  {psu.sby.current()}A")
print(
    f"PSU inventory: {psu.inventory.vendor()} {psu.inventory.model()} {psu.inventory.revision()} {psu.inventory.serial()}"
)

print(f"SRF MUX bus num {srf.mux.get_i2c_bus_number()}")
print(f"SRF GPIO bus num {srf.mux.ch00.gpio.get_i2c_bus_number()}")



from pathlib import Path

from ekfsm.core.components import HWModule
from ekfsm.core.sysfs import SysfsDevice, sysfs_root

from .generic import Device


class SMBIOS(Device):
    """
    A class to represent the SMBIOS device.

    A SMBIOS device is a virtual device that is used to read system
    configuration values from the DMI table.

    Note:
    Currently, only the board version / revision is read from the DMI table.
    """

    def __init__(
        self,
        name: str,
        parent: HWModule | None = None,
        children: list["Device"] | None = None,
        abort: bool = False,
        *args,
        **kwargs,
    ):
        self.sysfs_device: SysfsDevice | None = SysfsDevice(sysfs_root() / Path("devices/virtual/dmi/id"), False)

        super().__init__(name, parent, None, abort, *args, **kwargs)

    def revision(self) -> str:
        """
        Get the board revision from the DMI table.

        Returns
        -------
        str
            The board revision.
        """
        return self.sysfs.read_utf8("board_version")



import struct
from enum import Enum
from typing import Any, Tuple

from smbus2 import SMBus

from ekfsm.core.components import SysTree

from ..exceptions import AcquisitionError
from ..lock import Locker
from .generic import Device
from .imu import ImuSample


class CcuCommands(Enum):
    NOP = 0x01
    IMU_SAMPLES = 0x10
    FAN_STATUS = 0x11
    VIN_VOLTAGE = 0x12
    CCU_TEMPERATURE = 0x13
    CCU_HUMIDITY = 0x14
    PUSH_TEMPERATURE = 0x15
    SW_SHUTDOWN = 0x16
    WD_TRIGGER = 0x17
    IDENTIFY_FIRMWARE_TITLE = 0x80
    IDENTIFY_FIRMWARE_VERSION = 0x81
    LOAD_FIRMWARE_CHUNK = 0x82
    LOAD_PARAMETERSET = 0x83
    GET_PARAMETERSET_BEGIN = 0x84
    GET_PARAMETERSET_FOLLOW = 0x85
    RESTART = 0x8F


class EKFCcuUc(Device):
    """
    A class to communicate with I2C microcontroller on the EKF CCU.
    """

    def __init__(
        self,
        name: str,
        parent: SysTree | None,
        children: list["Device"] | None = None,
        abort: bool = False,
        debug: Any = None,  # XXX: What is this?
        *args,
        **kwargs,
    ):
        super().__init__(name, parent, None, abort, *args, **kwargs)
        self._i2c_addr = self.get_i2c_chip_addr()
        self._i2c_bus = self.get_i2c_bus_number()
        self._smbus = SMBus(self._i2c_bus)

    def __str__(self) -> str:
        return (
            f"EKFCCU - I2C Bus/Address: {self._i2c_bus}/{hex(self._i2c_addr)}; "
            f"sysfs_path: {self.sysfs_device.path if self.sysfs_device else ''}"
        )

    def temperature(self) -> float:
        """
        Get the temperature from the CCU thermal/humidity sensor.

        Note
        ----
        The CCU reads the temperature once per second.

        Returns
        -------
        float
            The temperature in degrees Celsius.

        Raises
        ------
        AcquisitionError
            If the temperature cannot be read, for example, because the sensor is not working.
        """
        return self._get_signed_word_data(CcuCommands.CCU_TEMPERATURE.value, "temperature") / 10.0

    def humidity(self) -> float:
        """
        Get the relative humidity from the CCU thermal/humidity sensor.
        The humidity is read once per second.

        Returns
        -------
        float
            The relative humidity in percent.

        Raises
        ------
        AcquisitionError
            If the humidity cannot be read, for example, because the sensor is not working.
        """
        return self._get_signed_word_data(CcuCommands.CCU_HUMIDITY.value, "humidity") / 10.0

    def vin_voltage(self) -> float:
        """
        Get the system input voltage from the CCU (the pimary voltage of the PSU).
        The voltage is read every 100ms.

        Returns
        -------
        float
            The system input voltage in volts.

        Raises
        ------
        AcquisitionError
            If the voltage cannot be read, for example, because the ADC is not working.
        """
        return self._get_signed_word_data(CcuCommands.VIN_VOLTAGE.value, "VIN voltage") / 10.0

    def _get_signed_word_data(self, cmd: int, what: str) -> int:
        v = self._smbus.read_word_data(self._i2c_addr, cmd)

        if v == 0x8000:
            raise AcquisitionError(f"Cannot read {what}")

        return struct.unpack("<h", struct.pack("<H", v))[0]

    def fan_status(self, fan: int) -> Tuple[float, float, int]:
        """
        Get the status of a fan.

        Parameters
        ----------
        fan
            The fan number (0-2).

        Returns
        -------
        desired: float
            The desired speed.
        actual: float
            The actual speed.
        diag: int
            The diagnostic value.

        Note
        ----
        The diagnostic value is a bitfield with the following meaning:

            - bit 0: 0 = fan status is invalid, 1 = fan status is valid
            - bit 1: 0 = no error detected, 1 = fan is stuck
        """
        data = self._smbus.read_block_data(self._i2c_addr, CcuCommands.FAN_STATUS.value)
        _data = bytes(data)
        desired, actual, diag = struct.unpack("<HHB", _data[fan * 5 : fan * 5 + 5])
        return desired, actual, diag

    def push_temperature(self, fan: int, temp: float) -> None:
        """
        Tell FAN controller the external temperature, usually the CPU temperature.

        Parameters
        ----------
        fan
            The fan number (0-2), or -1 to set the external temperature of all fans.

        temp
            The external temperature in degrees Celsius.

        Important
        ---------
        If push_temperature is no more called for a certain time (configurable with `fan-push-tout` parameter),
        the fan controller will fallback to it's default fan speed (configurable with the `fan-defrpm` parameter).

        """
        if fan == -1:
            fan = 0xFF
        data = struct.pack("<Bh", fan, int(temp * 10))
        self._smbus.write_block_data(self._i2c_addr, CcuCommands.PUSH_TEMPERATURE.value, list(data))

    def imu_sample(self) -> Tuple[ImuSample | None, bool]:
        """
        Read the next IMU sample from the CCU's IMU sample FIFO.

        If no sample is available, this method returns None.
        The second return value indicates if more samples are available in the FIFO.

        The CCU periodically samples the accelerometer and gyroscope data from the IMU and
        places it into a FIFO of 256 entries.
        Application must periodically read the samples from the FIFO to avoid overflow.

        FIFO overflow is indicated in the sample by the `lost` attribute.

        Note that the x, y, and z axes of the accelerometer and gyroscope
        are aligned to the mounting of the IMU on the CCU board. Please correct the axes if necessary
        to match the orientation of the IMU in your application.

        Returns
        -------
        imu_data: ImuSample | None
            The IMU sample, or None if no sample is available.
        more_samples: bool
            True if more samples are available in the FIFO, False otherwise.
        """
        more_samples = False
        _data = self._smbus.read_block_data(self._i2c_addr, CcuCommands.IMU_SAMPLES.value)
        data = bytes(_data)
        if len(data) < 14:
            return None, False  # No data available
        diag, fsr, acc_x, acc_y, acc_z, gyro_x, gyro_y, gyro_z = struct.unpack("<BBhhhhhh", data)
        imu_data = ImuSample(
            [
                self._scale_imu_accel(acc_x, fsr),
                self._scale_imu_accel(acc_y, fsr),
                self._scale_imu_accel(acc_z, fsr),
            ],
            [
                self._scale_imu_gyro(gyro_x, fsr),
                self._scale_imu_gyro(gyro_y, fsr),
                self._scale_imu_gyro(gyro_z, fsr),
            ],
            True if diag & 1 else False,
        )

        more_samples = True if (diag & 2 != 0) else False

        return imu_data, more_samples

    @staticmethod
    def _scale_imu_accel(val: int, fsr: int) -> float:
        fsr = fsr & 0xF
        scale = 16 / (1 << fsr)
        return val * (scale / 32768) * 9.80665  # convert to m/s^2

    @staticmethod
    def _scale_imu_gyro(val: int, fsr: int) -> float:
        fsr = fsr >> 4 & 0xF
        scale = 2000 / (1 << fsr)
        return val * (scale / 32768)

    def sw_shutdown(self) -> None:
        """
        Tell CCU that the system is going to shutdown.
        This cause the CCU's system state controller to enter shutdown state and power off the system after a certain time
        (parameter `shutdn-delay`).

        """
        self._smbus.write_byte(self._i2c_addr, CcuCommands.SW_SHUTDOWN.value)

    def wd_trigger(self) -> None:
        """
        Trigger the CCU's application watchdog.
        This will reset the watchdog timer.

        The CCU watchdog is only enabled when the parameter `wd-tout` is set to a value greater than 0. Triggering
        the watchdog when the timeout is 0 will have no effect.

        If the watchdog is not reset within the timeout, the CCU will power cycle the system.
        """
        self._smbus.write_byte(self._i2c_addr, CcuCommands.WD_TRIGGER.value)

    #
    # Management commands
    #
    def identify_firmware(self) -> Tuple[str, str]:
        """
        Get the firmware title and version of the CCU.

        Returns
        -------
        title: str
            The firmware title.
        version: str
            The firmware version.
        """
        title = bytes(self._smbus.read_block_data(self._i2c_addr, CcuCommands.IDENTIFY_FIRMWARE_TITLE.value)).decode(
            "utf-8"
        )
        version = bytes(
            self._smbus.read_block_data(self._i2c_addr, CcuCommands.IDENTIFY_FIRMWARE_VERSION.value)
        ).decode("utf-8")
        return title, version

    def load_firmware(self, firmware: bytes, progress_callback=None) -> None:
        """
        Load firmware into the CCU.

        The firmware must be the binary firmware file containing the application partition,
        typically named `fw-ccu-mm-default.bin`,
        where `mm` is the major version of the CCU hardware.

        The download can take several minutes, that is why a progress callback can be provided.

        When the download is complete and successful, the CCU will restart. To check if the firmware was loaded successfully,
        call :meth:`identify_firmware()` after the restart.

        Parameters
        ----------
        firmware
            The firmware binary data.

        progress_callback
            A callback function that is called with the current progress in bytes.

        Example
        -------
        >>> from ekfsm.devices import EkfCcuUc
        >>> ccu = EkfCcuUc("ccu")
        >>> firmware = open("fw-ccu-1.0.0.bin", "rb").read()
        >>> # Load firmware with progress callback
        >>> ccu.load_firmware(firmware, progress_callback=lambda x: print(f"Progress: {x} bytes"))
        """
        with Locker(self.name + "-load_firmware").lock():
            offset = 0
            max_chunk_len = 28

            while len(firmware) > 0:
                chunk, firmware = firmware[:max_chunk_len], firmware[max_chunk_len:]
                self._load_firmware_chunk(offset, len(firmware) == 0, chunk)
                offset += len(chunk)

                if len(firmware) != 0:
                    self._nop()

                if progress_callback is not None:
                    progress_callback(offset)

    def _load_firmware_chunk(self, offset: int, is_last: bool, data: bytes) -> None:
        if is_last:
            offset |= 0x80000000

        hdr = struct.pack("<I", offset)
        data = hdr + data

        self._smbus.write_block_data(self._i2c_addr, CcuCommands.LOAD_FIRMWARE_CHUNK.value, list(data))

    def get_parameterset(self) -> str:
        """
        Get the CCU parameterset in JSON format.

        A typical parameterset looks like this:

        .. code-block:: json

            {
                "version":      "factory",
                "parameters":   {
                        "num-fans":     "2",
                        "fan-temp2rpm": "25:2800;50:5000;100:6700",
                        "fan-rpm2duty": "2800:55;5000:88;6700:100",
                        "fan-defrpm":   "5500",
                        "fan-ppr":      "2",
                        "fan-push-tout":        "4000",
                        "pon-min-temp": "-25",
                        "pon-max-temp": "70",
                        "shutdn-delay": "120",
                        "wd-tout":      "0",
                        "pwrcycle-time":        "10"
                },
                "unsupported_parameters":       [],
                "missing_parameters":   ["num-fans", "fan-temp2rpm", "fan-rpm2duty", "fan-defrpm", "fan-ppr", \
                    "fan-push-tout", "pon-min-temp", "pon-max-temp", "shutdn-delay", "wd-tout", "pwrcycle-time"],
                "invalid_parameters":   [],
                "reboot_required":      false
            }

        `version` is the version of the parameterset. If no parameterset has been loaded by the user, the version is `factory`,
        otherwise it is the version of the loaded parameterset.

        `parameters` contains the current values of all parameters of the parameterset.

        `unsupported_parameters` contains the names of parameters that might have been downloaded, but
        are not supported by the CCU firmware.

        `missing_parameters` contains the names of parameters that have not been downloaded yet. Those parameters will
        have their default values.

        `invalid_parameters` contains the names of parameters that have been downloaded, but have invalid values.
        Those parameters will have their default values.

        `reboot_required` is a flag that indicates if a reboot is required to apply the parameterset.


        Returns
        -------
        str
            The parameterset in JSON format.

        """
        with Locker(self.name + "-parameterset").lock():
            json = b""
            begin = True

            while True:
                chunk = self._get_parameterset_chunk(begin)
                if len(chunk) < 32:
                    break
                # if chunk ends with zero byte, remove it (workaround for I2C slave bug)
                if chunk[-1] == 0:
                    chunk = chunk[:-1]
                json += chunk
                begin = False

            return json.decode("utf-8")

    def _get_parameterset_chunk(self, begin: bool) -> bytes:
        data = self._smbus.read_block_data(
            self._i2c_addr,
            (CcuCommands.GET_PARAMETERSET_BEGIN.value if begin else CcuCommands.GET_PARAMETERSET_FOLLOW.value),
        )
        return bytes(data)

    def load_parameterset(self, _cfg: str) -> None:
        """
        Load a parameterset into the CCU.

        The parameterset must be a JSON string containing the parameterset, for example:

        .. code-block:: json

            {
                "version": "1.0.0",
                "parameters":   {
                    "fan-defrpm": "6000"
                }
            }


        This would load a parameterset with just one parameter, the default fan speed. All other parameters will
        be set to their default values.

        Important
        ---------
        In order to apply the parameterset, the CCU must be restarted.

        Parameters
        ----------
        _cfg
            The parameterset in JSON format.

        Example
        -------
        >>> from ekfsm.devices import EkfCcuUc
        >>> ccu = EkfCcuUc("ccu")
        >>> # Load parameterset
        >>> ccu.load_parameterset('{"version": "1.0.0", "parameters": {"fan-defrpm": "6000"}}')
        >>> # Restart CCU to apply parameterset
        >>> ccu.restart()
        """
        with Locker(self.name + "-parameterset").lock():
            cfg = _cfg.encode("utf-8")
            offset = 0
            max_chunk_len = 28

            while len(cfg) > 0:
                chunk, cfg = cfg[:max_chunk_len], cfg[max_chunk_len:]
                self._load_parameterset_chunk(offset, len(cfg) == 0, chunk)
                offset += len(chunk)
                self._nop()

    def _load_parameterset_chunk(self, offset: int, is_last: bool, data: bytes) -> None:
        if is_last:
            offset |= 0x80000000

        hdr = struct.pack("<I", offset)
        data = hdr + data

        self._smbus.write_block_data(self._i2c_addr, CcuCommands.LOAD_PARAMETERSET.value, list(data))

    def restart(self) -> None:
        """
        Restart the CCU.
        """
        self._smbus.write_byte(self._i2c_addr, CcuCommands.RESTART.value)

    def _nop(self) -> None:
        self._smbus.read_word_data(self._i2c_addr, CcuCommands.NOP.value)



import threading
from ekfsm.devices.button import Button
from ekfsm.devices.generic import Device
from ekfsm.devices.io4edge import IO4Edge
import io4edge_client.binaryiotypeb as binio
import io4edge_client.functionblock as fb


class ButtonArray(Device):
    """
    Device class for handling an io4edge button array.

    To read button events, call the `read` method in a separate thread.

    Note
    ----
        Button handlers are called in the context of the `read` method's thread and need to be set in the Button instances.
    """

    def __init__(
        self,
        name: str,
        parent: IO4Edge,
        children: list[Device] | None = None,
        abort: bool = False,
        service_suffix: str | None = None,
        keepaliveInterval: int = 10000,
        *args,
        **kwargs,
    ):

        super().__init__(name, parent, children, abort, *args, **kwargs)

        self.name = name

        if service_suffix is not None:
            self.service_suffix = service_suffix
        else:
            self.service_suffix = name

        self.service_addr = f"{parent.deviceId}-{self.service_suffix}"

        self.client = binio.Client(
            self.service_addr, command_timeout=keepaliveInterval / 1000 + 5
        )

        self.subscriptionType = binio.Pb.SubscriptionType.BINARYIOTYPEB_ON_RISING_EDGE
        self.stream_cfg = fb.Pb.StreamControlStart(
            bucketSamples=1,  # 1 sample per bucket, also ein event pro bucket
            keepaliveInterval=keepaliveInterval,
            bufferedSamples=2,  # 2 samples werden gepuffert
            low_latency_mode=True,  # schickt soweit moeglich sofort die Events
        )

    def read(self, stop_event: threading.Event | None = None, timeout: float = 0.1):
        """
        Read all button events and dispatch to handlers.

        Parameters
        ----------
        stop_event : threading.Event, optional
            Event to signal stopping the reading loop. If None, the loop will run indefinitely.
        timeout : float, optional
            Timeout for reading from the stream in seconds. Default is 0.1 seconds.

        Note
        ----
            This method blocks and should be run in a separate thread.
        """
        self.client.start_stream(
            binio.Pb.StreamControlStart(
                subscribeChannel=tuple(
                    binio.Pb.SubscribeChannel(
                        channel=button.channel_id,
                        subscriptionType=self.subscriptionType,
                    )
                    for button in self.children
                    if isinstance(button, Button)
                )
            ),
            self.stream_cfg,
        )
        try:
            while not (stop_event and stop_event.is_set()):
                try:
                    _, samples = self.client.read_stream(timeout=timeout)
                    for sample in samples.samples:
                        for button in self.children:
                            if isinstance(button, Button):
                                pressed = bool(sample.inputs & (1 << button.channel_id))
                                if pressed and button.handler:
                                    button.handler()
                except TimeoutError:
                    continue
        finally:
            try:
                if stop_event:
                    stop_event.clear()
                self.client.stop_stream()
            except TimeoutError:
                pass

    def __repr__(self):
        return f"{self.name}; Service Address: {self.service_addr}"



import logging
import pprint
from pathlib import Path

from ekfsm.system import System

config = Path(__file__).parent / "zip2-ccu-only.yaml"
system = System(config)

logging.basicConfig(level=logging.DEBUG, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")

pprint.pprint(f"System slots {system.slots}")

system.print()

cpu = system["CPU1"]
cpuB = system.cpu1
cpuC = system[0]

assert cpu == cpuB == cpuC

cpu_slot = system.slots["SYSTEM_SLOT"]
cpu_slotB = system.slots.SYSTEM_SLOT
cpu_slotC = system.slots[0]

assert cpu_slot == cpu_slotB == cpu_slotC


cpu.print()
print(f"probing CPU: {cpu.probe()}")
print(f"inventory: {cpu.inventory.vendor()} {cpu.inventory.model()} {cpu.inventory.serial()}")

ccu = system[1]
ccu.print()
print(f"probing CCU: {ccu.probe()}")

eeprom = ccu.mux.ch00.eeprom

eeprom.write_cserial(12345678)
assert eeprom.cserial() == 12345678

eeprom.write_cmodel("SRS-C001")
assert eeprom.cmodel() == "SRS-C001"

eeprom.write_cvendor("EKF Elektronik")
assert eeprom.cvendor() == "EKF Elektronik"

eeprom.write_crevision("2.0")
assert eeprom.crevision() == "2.0"

eeprom.write_unit(1)
assert eeprom.unit() == 1



import os
import re
from abc import ABC, abstractmethod
from pathlib import Path

import gpiod
from gpiod.line import Direction, Value
from gpiod.line_settings import LineSettings
from more_itertools import first_true

from ekfsm.core.components import SysTree
from ekfsm.exceptions import GPIOError
from ekfsm.log import ekfsm_logger

from ..core.probe import ProbeableDevice
from .generic import Device

gpio_pat = re.compile(r"gpiochip\d+")


def get_gpio_major_minor(path: Path) -> tuple[int, int]:
    for d in path.iterdir():
        if gpio_pat.match(d.name):
            dev = d / "dev"
            if dev.exists():
                content = dev.read_text().strip()
                try:
                    major, minor = map(int, content.split(":"))
                except Exception as e:
                    raise GPIOError(
                        GPIOError.ErrorType.NO_MAJOR_MINOR,
                        f"No minor/major number found for GPIO device at {path}",
                    ) from e
                return major, minor

    raise GPIOError(
        GPIOError.ErrorType.NO_MAJOR_MINOR,
        f"No minor/major number found for GPIO device at {path}",
    )


def find_gpio_dev_with_major_minor(major: int, minor: int) -> Path | None:
    for dev in Path("/dev").iterdir():
        if gpio_pat.match(dev.name):
            stat_info = dev.stat()

            cmaj = os.major(stat_info.st_rdev)
            cmin = os.minor(stat_info.st_rdev)

            if cmaj == major and cmin == minor:
                return dev

    raise GPIOError(
        GPIOError.ErrorType.NO_MATCHING_DEVICE,
        f"Failed to find GPIO device with major {major} and minor {minor}",
    )


class GPIO(Device):
    def __init__(
        self,
        name: str,
        parent: SysTree | None = None,
        children: Device | None = None,
        abort: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(
            name,
            parent,
            None,
            abort,
            *args,
            **kwargs,
        )
        self.logger = ekfsm_logger("GPIODevice" + name)
        major, minor = self._find_gpio_dev(parent, *args, **kwargs)
        self.gpio = find_gpio_dev_with_major_minor(major, minor)

        assert self.gpio is not None
        match = re.search(r"\d+", self.gpio.name)

        if not match:
            raise GPIOError(GPIOError.ErrorType.NO_MATCHING_DEVICE, "Failed to find matching device")

        self.number: int = int(match.group().strip())

        self.init_dev()

    def _find_gpio_dev(self, parent: SysTree | None = None, *args, **kwargs) -> tuple[int, int]:
        self.addr = self.get_i2c_chip_addr()
        self.logger.debug(f"GPIO: {self.addr}")
        self.sysfs_device = self.get_i2c_sysfs_device(self.addr)
        return get_gpio_major_minor(self.sysfs_device.path)

    def init_dev(self):
        if self.gpio:
            try:
                self.dev = gpiod.Chip(str(self.gpio))
                self.initialized = True
                return
            except FileNotFoundError:
                self.initialized = False
                raise FileNotFoundError(f"{self.gpio} does not exist")

        self.initialized = False

    def num_lines(self) -> int:
        """
        Get number of GPIO lines available on the device.
        """
        return self.dev.get_info().num_lines

    def set_pin(self, pin: int, value: bool) -> None:
        """
        Set the value of a GPIO pin.

        Parameters
        ----------
        pin : int
            The pin number.
        value : bool
            The value to set.
        """
        v = Value.ACTIVE if value else Value.INACTIVE
        with self.dev.request_lines(
            consumer="set-pin",
            config={pin: gpiod.LineSettings()},
        ) as request:
            request.set_value(pin, v)

    def get_pin(self, pin: int) -> bool:
        """
        Get the value of a GPIO pin.

        Parameters
        ----------
        pin : int
            The pin number.

        Returns
        -------
        bool
            The value of the pin.
        """
        with self.dev.request_lines(
            consumer="get-pin",
            config={pin: gpiod.LineSettings()},
        ) as req:
            value = req.get_value(pin)
            return value == Value.ACTIVE

    def get_lines(self, lines: list[int]):
        if (
            invalid_pin := first_true(
                lines,
                pred=lambda line: line < 0 or line >= self.num_lines(),
                default=None,
            )
        ) is not None:
            raise GPIOError(GPIOError.ErrorType.INVALID_PIN, f"GPIO {invalid_pin} is invalid.")

    def set_direction(self, pin: int, direction: bool) -> None:
        """
        Set the direction of a GPIO pin.

        Parameters
        ----------
        pin : int
            The pin number.
        direction : bool
            The direction to set. True for output, False for input.
        """
        dir = Direction.OUTPUT if direction else Direction.INPUT
        self.dev.request_lines(
            consumer="set-direction",
            config={pin: LineSettings(direction=dir)},
        )

    def __str__(self) -> str:
        return (
            f"GPIO - Number: {self.number}; "
            f"sysfs_path: {self.sysfs_device.path if self.sysfs_device else ''} "
            f"(dev: {self.gpio if self.gpio else 'No matching device found'})"
        )


class GPIOExpander(GPIO):
    def __init__(
        self,
        name: str,
        parent: SysTree | None,
        children: Device | None = None,
        abort: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(name, parent, None, abort, *args, **kwargs)

    def __str__(self) -> str:
        return (
            f"GPIOExpander - Number: {self.number}; "
            f"sysfs_path: {self.sysfs_device.path if self.sysfs_device else ''}"
        )


class EKFIdentificationIOExpander(GPIOExpander, ProbeableDevice):
    def __init__(
        self,
        name: str,
        parent: SysTree | None,
        children: Device | None = None,
        abort: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(name, parent, None, abort, *args, **kwargs)

    def probe(self, *args, **kwargs) -> bool:
        from ekfsm.core import HWModule

        assert isinstance(self.hw_module, HWModule)
        id, _ = self.read_board_id_rev()
        self.logger.debug(f"Probing EKFIdentificationIOExpander: {id}")

        return self.hw_module.id == id

    def read_board_id_rev(self) -> tuple[int, int]:
        for pin in range(6, 8):
            self.set_direction(pin, True)
            self.set_pin(pin, False)
        for pin in range(0, 6):
            self.set_direction(pin, False)

        v_gnd = self.read_id_gpio_inputs()
        self.set_pin(7, True)
        v_7h = self.read_id_gpio_inputs()
        self.set_pin(6, True)
        v_6h = self.read_id_gpio_inputs()

        readings = [v_gnd, v_7h, v_6h]
        bit_sums = []
        for bit in range(5, -1, -1):
            s = 0
            for reading in readings:
                s += (reading >> bit) & 1
            bit_sums.append(s)

        return (
            sum(val * (4**i) for i, val in enumerate(reversed(bit_sums[2:]))),
            self._get_board_rev(bit_sums[:2]),  # board_rev
        )

    def revision(self) -> str:
        _, rev = self.read_board_id_rev()
        return str(rev)

    def read_id_gpio_inputs(self) -> int:
        value = 0
        for pin in range(6):
            if self.get_pin(pin):
                value |= 1 << pin

        return value

    @staticmethod
    def _get_board_rev(bits: list[int]) -> int:
        """Convert 2-bit sum values to board revision number."""
        rev_map = {
            (0, 0): 0,
            (0, 1): 1,
            (0, 2): 2,
            (1, 0): 3,
            (1, 1): 4,
            (1, 2): 5,
            (2, 0): 6,
            (2, 1): 7,
            (2, 2): 8,
        }
        return rev_map.get((bits[0], bits[1]), -1)

    def __str__(self) -> str:
        return (
            f"EKFIdentificationIOExpander - Number: {self.number}; "
            f'sysfs_path: {self.sysfs_device.path if self.sysfs_device else ""}'
        )


class SimGpio(ABC):
    @abstractmethod
    def num_lines(self) -> int:
        pass

    @abstractmethod
    def set_pin(self, pin: int, value: bool) -> None:
        pass

    @abstractmethod
    def get_pin(self, pin: int) -> bool:
        pass

    @abstractmethod
    def set_direction(self, pin: int, direction: bool) -> None:
        pass


class EKFIdSimGpio(SimGpio):
    def __init__(self, coding_gnd, coding_vcc, coding_6, coding_7) -> None:
        self._coding_gnd = coding_gnd
        self._coding_vcc = coding_vcc
        self._coding_6 = coding_6
        self._coding_7 = coding_7
        self._dir = 0
        self._out = 0
        self._in = 0

    def num_lines(self) -> int:
        return 8

    def set_pin(self, pin: int, value: bool) -> None:
        mask = 1 << pin
        if value not in [0, 1]:
            raise RuntimeError("value must be 0 or 1")
        if self._dir & mask:
            self._out = (self._out & ~mask) | (value << pin)
        else:
            raise RuntimeError("pin not set as output")

    def get_pin(self, pin: int) -> bool:
        mask = 1 << pin
        if self._coding_gnd & mask:
            return False
        if self._coding_vcc & mask:
            return True
        if self._coding_6 & mask:
            return True if self._out & (1 << 6) else False
        if self._coding_7 & mask:
            return True if self._out & (1 << 7) else False
        return False

    def set_direction(self, pin: int, direction: bool) -> None:
        if direction == 1 and (pin != 6 and pin != 7):
            raise RuntimeError("only pins 6 and 7 supported as output")
        mask = 1 << pin
        if direction:
            self._dir |= mask
        else:
            self._dir &= ~mask



from pathlib import Path
from typing import TYPE_CHECKING, Callable

from munch import Munch

from ekfsm.core.components import SysTree
from ekfsm.core.sysfs import SysfsDevice, sysfs_root
from ekfsm.exceptions import ConfigError, SysFSError, UnsupportedModeError

if TYPE_CHECKING:
    from ekfsm.core.components import HWModule


class Device(SysTree):
    """
    A generic device.
    """

    def __init__(
        self,
        name: str,
        parent: SysTree | None = None,
        children: list["Device"] | None = None,
        abort: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(name, abort=abort)
        self.parent = parent
        self.device_args = kwargs

        if children:
            self.children = children

        # Needs to be set during init because root will be changed after tree is complete
        self.hw_module = self.root  # pyright: ignore[reportAttributeAccessIssue]

        # I2C initialization
        if not hasattr(self, "sysfs_device") or self.sysfs_device is None:
            self.sysfs_device: SysfsDevice | None = None

        # Post-initialization steps
        self._provides_attrs = kwargs.get("provides", {})
        self.provides = self.__post_init__(Munch(self._provides_attrs), abort)

    def __post_init__(self, provides: Munch, abort) -> Munch:
        for key, fields in provides.items():
            if isinstance(fields, dict):
                provides[key] = self.__post_init__(Munch(fields), abort)
            elif isinstance(fields, list | str):
                provides[key] = Munch()

                if isinstance(fields, str):
                    fields = [fields]

                while fields:
                    interface = fields.pop()

                    # TODO: Check if this is superfluous after schema validation
                    if isinstance(interface, dict):
                        name = list(interface.keys())[0]

                        try:
                            func = list(interface.values())[0]
                        except IndexError:
                            raise ConfigError(f"{self.name}: No function given for interface {name}.")

                        if not hasattr(self, func):
                            if abort:
                                raise NotImplementedError(
                                    f"{self.name}: Function {func} for interface {name} not implemented."
                                )
                            continue

                        provides[key].update({name: getattr(self, func)})
                    else:
                        if not hasattr(self, interface):
                            if abort:
                                raise NotImplementedError(
                                    f"{self.name}: Function {interface} for provider {key} not implemented."
                                )
                            continue

                        provides[key].update({interface: getattr(self, interface)})

        return provides

    @property
    def sysfs(self):
        """
        Access sysfs device for device

        Returns
        -------
        Sysfs device for device

        Raises
        ------
        SysFSError
            If the SysFSDevice does not exist
        """
        if self.sysfs_device is None:
            raise SysFSError(f"No sysfs device attached to device {self.name}")

        return self.sysfs_device

    def read_attr(self, attr: str, mode: str = "utf", strip: bool = True):
        match mode:
            case "utf":
                return self.sysfs.read_utf8(attr, strip)
            case "bytes":
                return self.sysfs.read_bytes(attr)
            case "float":
                return self.sysfs.read_float(attr)
            case "int":
                return self.sysfs.read_int(attr)
            case _:
                raise UnsupportedModeError(f"Mode {mode} is not supported")

    def read_attr_or_default(self, attr: str, mode: str = "utf", strip: bool = True, default=None):
        try:
            return self.read_attr(attr, mode, strip)
        except UnsupportedModeError:
            raise
        except Exception:
            if default is not None:
                return default
            return None

    def read_sysfs_bytes(self, attr) -> bytes:
        if len(attr) <= 0:
            raise SysFSError("sysfs attribute name too short")

        return self.sysfs.read_bytes(attr)

    def read_sysfs_attr_bytes(self, attr: str) -> bytes | None:
        """
        Read a sysfs attribute as bytes.

        Parameters
        ----------
        attr
            The sysfs attribute to read.

        Returns
        -------
        content: bytes
            The contents of the sysfs attribute as bytes.
        None:
            If the sysfs device is not set or the attribute does not exist.
        """
        if self.sysfs_device is not None and len(attr) != 0 and attr in [x.name for x in self.sysfs_device.attributes]:
            return self.sysfs_device.read_attr_bytes(attr)
        return None

    def read_sysfs_attr_utf8(self, attr: str) -> str | None:
        """
        Read a sysfs attribute as UTF-8 string.

        Parameters
        ----------
        attr
            The sysfs attribute to read.

        Returns
        -------
        content: str
            The contents of the sysfs attribute as UTF-8 string.
        None:
            If the sysfs device is not set or the attribute does not exist.
        """
        try:
            return self.sysfs.read_utf8(attr)
        except Exception:
            return None

    def write_sysfs_attr(self, attr: str, data: str | bytes) -> None:
        """
        Write data to a sysfs attribute.

        Parameters
        ----------
        attr
            The sysfs attribute to write to.
        data
            The data to write to the sysfs attribute.
        """
        if self.sysfs_device and len(attr) != 0:
            return self.sysfs_device.write_attr(attr, data)
        return None

    @property
    def hw_module(self) -> "HWModule":
        """
        Get or set the HWModule instance that this device belongs to.

        Parameters
        ----------
        hw_module: optional
            The HWModule instance to set.

        Returns
        -------
        HWModule
            The HWModule instance that this device belongs to.
        None
            If used as a setter.
        """
        from ekfsm.core.components import HWModule

        if isinstance(self._hw_module, HWModule):
            return self._hw_module
        else:
            raise RuntimeError("Device is not a child of HWModule")

    @hw_module.setter
    def hw_module(self, hw_module: "HWModule") -> None:
        self._hw_module = hw_module

    def get_i2c_chip_addr(self) -> int:
        assert self.parent is not None

        chip_addr = self.device_args.get("addr")
        if chip_addr is None:
            raise ConfigError(f"{self.name}: Chip address not provided in board definition")

        if not hasattr(self.parent, "sysfs_device") or self.parent.sysfs_device is None:
            # our device is the top level device of the slot
            # compute chip address from board yaml and slot attributes
            slot_attributes = self.hw_module.slot.attributes

            if slot_attributes is None:
                raise ConfigError(f"{self.name}: Slot attributes not provided in system configuration")

            if not self.hw_module.is_master:
                # slot coding is only used for non-master devices
                if not hasattr(slot_attributes, "slot_coding"):
                    raise ConfigError(f"{self.name}: Slot coding not provided in slot attributes")

                slot_coding_mask = 0xFF

                if hasattr(slot_attributes, "slot_coding_mask"):
                    slot_coding_mask = slot_attributes.slot_coding_mask

                chip_addr |= slot_attributes.slot_coding & slot_coding_mask

        return chip_addr

    def get_i2c_sysfs_device(self, addr: int, driver_required=True, find_driver: Callable | None = None) -> SysfsDevice:
        from ekfsm.core.components import HWModule

        parent = self.parent
        assert parent is not None

        # If parent is a HWModule, we can get the i2c bus from the master device
        # XXX: Does this still hold true after refactoring?
        if isinstance(parent, HWModule):
            i2c_bus_path = self.__master_i2c_bus()
        else:
            # otherwise the parent must be a MuxChannel
            from ekfsm.devices.mux import MuxChannel

            assert isinstance(parent, MuxChannel)
            assert parent.sysfs_device is not None
            i2c_bus_path = parent.sysfs_device.path

        # search for device with addr
        for entry in i2c_bus_path.iterdir():
            if (
                entry.is_dir()
                and not (entry / "new_device").exists()  # skip bus entries
                and (entry / "name").exists()
            ):
                # PRP devices unfortunately do not readily expose the underlying I2C address of the device like
                # regular I2C devices that follow the `${I2C_BUS}-${ADDR}` pattern. To address this issue, we
                # initialize the ACPI _STR object for each PRP device with the necessary information, which is
                # accessible in the `${DEVICE_SYSFS_PATH}/firmware_node/description` file.
                if (entry / "firmware_node").exists() and (entry / "firmware_node" / "description").exists():
                    description = (entry / "firmware_node/description").read_text().strip()
                    acpi_addr = int(description.split(" - ")[0], 16)

                    if acpi_addr == addr:
                        return SysfsDevice(entry, driver_required, find_driver)

                # For regular non-PRP devices, the address is contained in the directory name (e.g. 2-0018).
                else:
                    acpi_addr = int(entry.name.split("-")[1], 16)

                    if acpi_addr == addr:
                        return SysfsDevice(entry, driver_required, find_driver)

        raise FileNotFoundError(f"Device with address 0x{addr:x} not found in {i2c_bus_path}")

    @staticmethod
    def __master_i2c_get_config(master: "HWModule") -> dict:
        if master.config.get("bus_masters") is not None and master.config["bus_masters"].get("i2c") is not None:
            return master.config["bus_masters"]["i2c"]
        else:
            raise ConfigError("Master definition incomplete")

    def __master_i2c_bus(self) -> Path:
        if self.hw_module.is_master:
            # we are the master
            master = self.hw_module
            master_key = "MASTER_LOCAL_DEFAULT"
            override_master_key = self.device_args.get("i2c_master", None)

            if override_master_key is not None:
                master_key = override_master_key
        else:
            # another board is the master
            if self.hw_module.slot.master is None:
                raise ConfigError(f"{self.name}: Master board not found in slot attributes")

            master = self.hw_module.slot.master
            master_key = self.hw_module.slot.slot_type.name

        i2c_masters = self.__master_i2c_get_config(master)

        if i2c_masters.get(master_key) is not None:
            dir = sysfs_root() / Path(i2c_masters[master_key])
            bus_dirs = list(dir.glob("i2c-*"))

            if len(bus_dirs) == 1:
                return bus_dirs[0]
            elif len(bus_dirs) > 1:
                raise ConfigError(f"Multiple master I2C buses found for {master_key}")

            raise ConfigError(f"No master I2C bus found for {master_key}")
        else:
            raise ConfigError(f"Master I2C bus not found for {master_key}")

    def get_i2c_bus_number(self) -> int:
        """
        Get the I2C bus number of the device. Works for devices that do not have a sysfs_device attribute.
        """
        from ekfsm.devices.mux import MuxChannel

        if isinstance(self, MuxChannel):
            raise RuntimeError(f"{self.name}: MuxChannel does not have a bus number")

        if self.sysfs_device is None:
            if self.parent is None:
                raise RuntimeError(f"{self.name}: Must have a parent to get bus number")
            parent_path = self.parent.sysfs_device.path
        else:
            parent_path = self.sysfs_device.path.parent

        if parent_path.is_symlink():
            parent_path = parent_path.readlink()

        bus_number = parent_path.name.split("-")[1]

        return int(bus_number)

    def __repr__(self) -> str:
        sysfs_path = getattr(self.sysfs_device, "path", "")
        return f"{self.name}; Path: {sysfs_path}"



import logging

#
# We follow the recommendations from https://docs.python.org/3/howto/logging.html#configuring-logging-for-a-library
#
# By default, if the application does not configure logging, the logging module will log
# only messages with level WARNING or above and is using the default formatting, i.e.
# only the message is printed.
#
# To get a more verbose output, the application should call, for example
# logging.basicConfig(level=logging.DEBUG, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')


def ekfsm_logger(name: str) -> logging.Logger:
    """
    Create a logger with the name 'ekfsm:name'

    Returns
    -------
    logging.Logger
        The logger object.

    Parameters
    ----------
    name
        The name of the module, class or object that is using the logger.
    """
    return logging.getLogger("ekfsm:" + name)



import re
from enum import IntFlag
from functools import wraps
from time import sleep

from ekfsm.core.components import SysTree
from ekfsm.exceptions import HWMonError
from ekfsm.log import ekfsm_logger

from ..core.probe import ProbeableDevice
from ..core.sysfs import list_sysfs_attributes, sysfs_root
from .generic import Device

__all__ = ["PSUStatus", "PMBus", "retry"]

logger = ekfsm_logger(__name__)


def retry(max_attempts=5, delay=0.5):
    """
    Retry decorator.

    Decorator that retries a function a number of times before giving up.

    This is useful for functions that may fail due to transient errors.

    Note
    ----
    This is needed for certain PMBus commands that may fail due to transient errors
    because page switching timing is not effectively handled by older kernel versions.

    Important
    ---------
    This decorator is _not_ thread-safe across multiple ekfsm processes. Unfortunately,
    we cannot use fcntl or flock syscalls with files on virtual filesystems like sysfs.

    Parameters
    ----------
    max_attempts
        The maximum number of attempts before giving up.
    delay
        The delay in seconds between attempts.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            attempts = 0
            while attempts < max_attempts:
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    attempts += 1
                    if attempts == max_attempts:
                        logger.exception(f"Failed to execute {func.__name__} after {max_attempts} attempts: {e}")
                        raise e
                    logger.info(f"Retrying execution of {func.__name__} in {delay}s...")
                sleep(delay)

        return wrapper

    return decorator


class PSUStatus(IntFlag):
    """
    Represents the status of a PSU according to STATUS_BYTE register.

    See also
    --------
    External Documentation:
    `PMBus Power System Management Protocol Specification - Part II - Revision 1.4, Fig. 60 <https://pmbus.org/>`_

    Example
    -------
    >>> from ekfsm.devices.pmbus import PsuStatus
    >>> status = PsuStatus(0x1F)
    >>> status
    <PsuStatus.OUTPUT_OVERCURRENT|INPUT_UNDERVOLTAGE|TEMP_ANORMALY|COMMUNICATION_ERROR|ERROR: 31>
    >>> PsuStatus.OUTPUT_OVERCURRENT in status
    True
    >>> # OK is always present
    >>> PsuStatus.OK in status
    True
    >>> # Instead, check if status is OK
    >>> status == PsuStatus(0x00)
    False
    >>> PsuStatus.OUTPUT_OVERCURRENT in status
    False
    """

    OUTPUT_OVERVOLTAGE = 0x20
    OUTPUT_OVERCURRENT = 0x10
    INPUT_UNDERVOLTAGE = 0x08
    TEMP_ANORMALY = 0x04
    COMMUNICATION_ERROR = 0x02
    ERROR = 0x01
    OK = 0x00


logger = ekfsm_logger(__name__)


class PMBus(Device, ProbeableDevice):
    def __init__(
        self,
        name: str,
        parent: SysTree | None = None,
        children: list[Device] | None = None,
        abort: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(name, parent, children, abort, *args, **kwargs)
        self.addr = self.get_i2c_chip_addr()
        self.sysfs_device = self.get_i2c_sysfs_device(self.addr, driver_required=True)

        try:
            for entry in self.sysfs_device.path.glob("hwmon/hwmon*"):
                if entry.is_dir():
                    attrs = list_sysfs_attributes(entry)
                    self.sysfs_device.extend_attributes(attrs)

                    debug_attrs_path = sysfs_root().joinpath(f"kernel/debug/pmbus/{entry.name}")
                    debug_attrs = list_sysfs_attributes(debug_attrs_path)
                    self.sysfs_device.extend_attributes(debug_attrs)
        except FileNotFoundError:
            logger.debug("Expected sysfs attribute not found")
        except StopIteration:
            raise HWMonError("Device is not managed by hwmon subsystem")

    def probe(self, *args, **kwargs) -> bool:
        from ekfsm.core import HWModule

        assert isinstance(self.hw_module, HWModule)
        # compare the regexp from the board yaml file with the model
        return re.match(self.hw_module.id, self.model()) is not None

    # Voltage and Current Interfaces
    def __convert_and_scale(self, attr: str) -> float:
        return self.sysfs.read_float(attr) / 1000.0

    @retry()
    def in1_input(self) -> float:
        """
        Get input voltage of PSU page 1.

        Returns
        -------
            Input voltage in volts
        """
        return self.__convert_and_scale("in1_input")

    @retry()
    def in2_input(self) -> float:
        """
        Get input voltage of PSU page 2.

        Returns
        -------
            Input voltage in volts
        """
        return self.__convert_and_scale("in2_input")

    @retry()
    def curr1_input(self) -> float:
        """
        Get input current of PSU page 1.

        Returns
        -------
            Input current in amperes
        """
        return self.__convert_and_scale("curr1_input")

    @retry()
    def curr2_input(self) -> float:
        """
        Get input current of PSU page 2.

        Returns
        -------
            Input current in amperes
        """
        return self.__convert_and_scale("curr2_input")

    # Status Interface
    @retry()
    def status0_input(self) -> PSUStatus:
        """
        Get the status of PSU page 1.

        Returns
        -------
            PSU status as defined in PSUStatus
        """
        status = self.sysfs.read_int("status0_input")
        return PSUStatus(status)

    @retry()
    def status1_input(self) -> PSUStatus:
        """
        Get the status of PSU page 2.

        Returns
        -------
            PSU status as defined in PSUStatus
        """
        status = self.sysfs.read_int("status1_input")
        return PSUStatus(status)

    # Temperature Interface
    @retry()
    def temp1_input(self) -> float:
        """
        Get the PSU temperature.

        Returns
        -------
            PSU temperature in degrees celsius
        """
        return self.__convert_and_scale("temp1_input")

    # Inventory Interface
    def vendor(self) -> str:
        """
        Get the vendor of the PSU.

        Returns
        -------
            PSU vendor
        """
        return self.sysfs.read_utf8("vendor")

    def model(self) -> str:
        """
        Get the model of the PSU.

        Returns
        -------
            PSU model
        """
        return self.sysfs.read_utf8("model")

    def serial(self) -> str:
        """
        Get the serial number of the PSU.

        Returns
        -------
            PSU serial number
        """
        return self.sysfs.read_utf8("serial")

    def revision(self) -> str:
        """
        Get the revision of the PSU.

        Returns
        -------
            PSU revision
        """
        return self.sysfs.read_utf8("revision")



import logging
import pprint
from pathlib import Path

import pytest
import yaml

from ekfsm.core.utils import module_schema


def test_module_schema():
    boards_path = Path(__file__).parent.parent / "ekfsm" / "boards" / "oem" / "ekf"

    for module_path in boards_path.glob("*.yaml"):
        pprint.pprint(f"Module: {module_path}")

        with open(module_path) as file:
            yaml_data = yaml.safe_load(file)
            module_schema.validate(yaml_data)


if __name__ == "__main__":
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )

    pytest.main([__file__])



import pprint
import logging
from pathlib import Path

from ekfsm.system import System
from ekfsm.devices.colorLed import Color

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)

config = Path(__file__).parent / "cctv.yaml"
system = System(config, abort=True)

pprint.pprint(f"System slots {system.slots}")

system.print()

cpu = system["CPU"]
cpuB = system.cpu
cpuC = system[0]

assert cpu == cpuB == cpuC

# To check why below is failing
# cpu_slot = system.slots["SYSTEM_SLOT"]
# cpu_slotB = system.slots.SYSTEM_SLOT
# cpu_slotC = system.slots[0]

# assert cpu_slot == cpu_slotB == cpu_slotC

cpu.print()
print(f"probing CPU: {cpu.probe()}")
print(
    f"inventory: {cpu.inventory.vendor()} {cpu.inventory.model()} {cpu.inventory.serial()}"
)

smc = system.smc

smc.print()
smc.leds.led0.set(Color.RED, True)



from typing import Callable, Optional
from ekfsm.core.components import HWModule
from ekfsm.devices.generic import Device
import io4edge_client.core.coreclient as Client

from re import sub


class IO4Edge(Device):
    """
    Device class for handling IO4Edge devices.
    """

    def __init__(
        self,
        name: str,
        parent: HWModule | None = None,
        children: list[Device] | None = None,
        abort: bool = False,
        *args,
        **kwargs,
    ):

        super().__init__(name, parent, children, abort, *args, **kwargs)

        attr = self.hw_module.slot.attributes
        if (
            attr is None
            or not hasattr(attr, "slot_coding")
            or getattr(attr, "slot_coding") is None
        ):
            raise ValueError(
                f"Slot attributes for {self.hw_module.slot.name} are not set or do not contain 'slot_coding'."
            )
        else:
            geoaddr = int(attr.slot_coding)
            self._geoaddr = geoaddr

        _, module_name = sub(r"-.*$", "", self.hw_module.board_type).split(maxsplit=1)
        self._module_name = module_name
        self.client = Client.new_core_client(self.deviceId)

    @property
    def deviceId(self) -> str:
        """
        Returns the device ID for the IO4Edge device.
        The device ID is a combination of the module name and the geo address.
        """
        return f"{self._module_name}-geo_addr{self._geoaddr:02d}"

    def identify_firmware(self) -> tuple[str, str]:
        return (
            self.client.identify_firmware().title,
            self.client.identify_firmware().version,
        )

    def load_firmware(
        self, cfg: bytes, progress_callback: Optional[Callable[[float], None]] = None
    ) -> None:
        """
        Load firmware onto the IO4Edge device.

        cfg
            Firmware configuration bytes.
        progress_callback
            Optional callback for progress updates.
        """
        self.client.load_firmware(cfg, progress_callback)

    def restart(self) -> None:
        self.client.restart()

    def load_parameter(self, name: str, value: str) -> None:
        """
        Set a parameter onto the IO4Edge device.

        cfg
            The name of the parameter to load.
        value
            The value to set for the parameter.
        """
        self.client.set_persistent_parameter(name, value)

    def get_parameter(self, name: str) -> str:
        """
        Get a parameter value from the IO4Edge device.

        Returns
            The value of the requested parameter.
        """
        return self.client.get_persistent_parameter(name)

    def __repr__(self):
        return f"{self.name}; DeviceId: {self.deviceId}"



from ekfsm.devices.generic import Device
from ekfsm.devices.io4edge import IO4Edge
from io4edge_client.analogintypea import Client


class ThermalHumidity(Device):
    """
    Device class for handling a thermal humidity sensor.
    """

    def __init__(
        self,
        name: str,
        parent: IO4Edge,
        children: list[Device] | None = None,
        abort: bool = False,
        service_suffix: str | None = None,
        *args,
        **kwargs,
    ):

        super().__init__(name, parent, children, abort, *args, **kwargs)

        self.name = name

        if service_suffix is not None:
            self.service_suffix = service_suffix
        else:
            self.service_suffix = name

        self.service_addr = f"{parent.deviceId}-{self.service_suffix}"

        self.client = Client(self.service_addr)

    def temperature(self) -> float:
        """
        Get the temperature in Celsius.

        Raises
        ------
        RuntimeError
            if the command fails
        TimeoutError
            if the command times out
        """
        return self.client.value()

    def __repr__(self):
        return f"{self.name}; Service Address: {self.service_addr}"



"""
A module containing classes to represent EEPROM devices.

Routine Listings
----------------
    :py:class:`EEPROM`
    :py:class:`Validatable_EEPROM`
    :py:class:`EKF_EEPROM`
    :py:class:`validated`
"""

from abc import ABC, abstractmethod
from datetime import date
from functools import wraps
from typing import Any, Callable, Literal, Sequence

from hexdump import hexdump

from ekfsm.core.components import SysTree
from ekfsm.core.probe import ProbeableDevice
from ekfsm.exceptions import DataCorruptionError, DriverError, SysFSError
from ekfsm.log import ekfsm_logger

from .generic import Device
from .utils import get_crc16_xmodem

logger = ekfsm_logger(__name__)


def validated(func: Callable[..., Any]) -> Callable[..., Any]:
    """
    A decorator to validate the CRC of the EEPROM content before executing a method.

    Parameters
    ----------
    func
        The method to validate.

    Note
    ----
    This decorator should be used on methods that read data from an EEPROM.
    """

    @wraps(func)
    def validate(self, *args, **kwargs):
        logger.debug(f"Validating EEPROM content for {self.name}")
        if not self.valid:
            raise DataCorruptionError("CRC validation failed")
        logger.debug(f"EEPROM content is valid for {self.name}")
        return func(self, *args, **kwargs)

    return validate


class EEPROM(Device):
    """
    A class used to represent a generic EEPROM device.

    Parameters
    ----------
    name
        The name of the EEPROM device.
    parent
        The parent device of the EEPROM in the :py:class:`~.generic.Device` tree.


    Caution
    -------
    The following conditions must be met for this class to work properly:
        - EEPROM must be I2C accessable
        - EEPROM must have a sysfs device


    Note
    ----
    This class should be inherited by classes representing specific EEPROM devices and defining custom storage schemes.
    """

    def __init__(
        self,
        name: str,
        parent: SysTree | None = None,
        children: list[Device] | None = None,
        abort: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(name, parent, None, abort, *args, **kwargs)

        self.addr = self.get_i2c_chip_addr()
        self.sysfs_device = self.get_i2c_sysfs_device(self.addr)

        if not self.sysfs_device.get_driver():
            raise DriverError("No driver attached to device {self.name}")

        self._update_content()

    def _update_content(self) -> None:
        """
        Update the content of the EEPROM device.


        Note
        ----
            - This method should be called whenever the content of the EEPROM is updated (after each write op).
            - Inheriting classes should call this method before updating their own attributes.
        """
        logger.debug("Reading data")
        try:
            data = self.read()
            self._content = data
        except Exception as e:
            logger.error(f"Error reading data, {e}")
            self._content = b""

    def read(self) -> bytes:
        """
        Read the content of the EEPROM.

        Returns
        -------
            The content of the EEPROM.

        Raises
        ------
        SysFSError
            No sysfs device found for EEPROM or `eeprom` attribute does not exist
        """
        return self.read_sysfs_bytes("eeprom")

    def write(self, data: bytes, offset: int = 0) -> None:
        """
        Write data to the EEPROM.

        Parameters
        ----------
        data
            The data to write to the EEPROM.
        offset
            The offset at which to start writing the data.

        Raises
        ------
        RuntimeError
            If the sysfs device is not found.
        FileNotFoundError
            If the EEPROM sysfs file is not found.
        DataCorruptionError
            If an error occurs during the write operation.

        Note
        ----
        Operation is checked for data corruption by reading back the written data.

        Important
        ---------
        The offset parameter is only supported if the EEPROM driver is bin_attribute enabled.

        For almost any other native sysfs attribute, this is NOT the case!
        """

        if self.sysfs_device:
            attr = next(x for x in self.sysfs_device.attributes if x.name == "eeprom")
            logger.info(f"Writing {len(data)} bytes to EEPROM at offset {offset}")
            if attr.is_sysfs_attr() and data is not None:
                mode = "r+" if isinstance(data, str) else "rb+"
            try:
                with open(attr.path, mode) as f:
                    f.seek(offset)
                    f.write(data)
            except OSError as e:
                raise SysFSError("Error accessing SysFS attribute") from e
        else:
            raise RuntimeError("No sysfs device for EEPROM")

        self._update_content()
        written = self._content[offset : offset + len(data)]
        if not written == data:
            raise DataCorruptionError(
                "Error during EEPROM write, data is not the same as read back"
            )

    def print(self):
        hexdump(self._content)


class Validatable_EEPROM(EEPROM, ABC):
    """
    Abstract class used to represent an EEPROM device using CRC to validate its content.

    Parameters
    ----------
    crc_pos
        The position of the CRC value in the EEPROM content (`'start'` or `'end'`).
    crc_length
        The length of the CRC value in number of bytes (defaults to 2).

    Note
    ----
    - Derived classes must implement a method to compute the CRC value of the EEPROM content.
    - If the CRC position differs from the shipped schema `('start' | 'end')`,
      the derived class must override the :meth:`~Validatable_EEPROM._update_content` method.
    - Validity of individual content fields stored/returned by attributes, methods or properties
      can be achieved by using the :py:func:`validated` decorator.

    See Also
    --------
    Validatable_EEPROM._compute_crc : Method to compute the CRC value of the EEPROM content.
    """

    def __init__(
        self,
        name: str,
        parent: SysTree | None = None,
        children: list[Device] | None = None,
        abort: bool = False,
        crc_pos: Literal["start", "end"] = "end",
        crc_length: int = 2,
        *args,
        **kwargs,
    ) -> None:
        self._crc_length: int = crc_length
        self._crc_pos: str = crc_pos

        super().__init__(name, parent, children, abort, *args, **kwargs)

        self._crc_pos_start = len(self._data) if self._crc_pos == "end" else 0
        self._crc_pos_end = self._crc_pos_start + self._crc_length

    def _update_content(self) -> None:
        """
        Update the content of the EEPROM device (checksum excluded).
        """
        super()._update_content()

        # Firmware data without CRC
        self._data: bytes = (
            self._content[self._crc_length :]
            if self._crc_pos == "start"
            else self._content[: -self._crc_length :]
        )

    def _update_crc(self) -> None:
        """
        Update the CRC value of the EEPROM content.
        """
        self.crc = self._compute_crc()

    def _get_crc_value(self) -> int:
        return int.from_bytes(
            self._content[self._crc_pos_start : self._crc_pos_end], byteorder="little"
        )

    @property
    def crc(self) -> int:
        """
        Gets or sets the CRC value of the EEPROM content.

        Parameters
        ----------
        value: optional
            The CRC value to write to the EEPROM.

        Returns
        -------
            int
                The CRC value currently stored in the EEPROM if used as *getter*.
            None
                If used as *setter*.


        Caution
        -------
        The *setter* actually writes the CRC value to the EEPROM.


        Warning
        -------
        The *setter* method should be used with caution as it can lead to data corruption if the CRC value is not correct!


        Note
        ----
        The *setter* is usually triggered automatically after a successful write operation
        and in most cases, there is no need to call it manually.
        """
        return self._get_crc_value()

    @crc.setter
    def crc(self, value: int) -> None:
        crc_bytes = value.to_bytes(self._crc_length, byteorder="little")
        logger.debug(f"Writing CRC value {value} to EEPROM")

        try:
            super().write(crc_bytes, self._crc_pos_start)
        except Exception as e:
            logger.error(f"Error writing CRC value to EEPROM, {e}")

        # self._update_content()

    @property
    def valid(self) -> bool:
        """
        Checks if the EEPROM content is valid by comparing the stored CRC value with the computed CRC value.

        Returns
        -------
        bool
            `True` if the EEPROM content is valid, `False` otherwise.
        """
        return self._compute_crc() == self.crc

    @abstractmethod
    def _compute_crc(self) -> int:
        """
        This method should be implemented by derived classes to compute the CRC value of the EEPROM content.

        Returns
        -------
            The computed CRC value.
        """
        pass

    def write(self, data: bytes, offset: int = 0) -> None:
        try:
            super().write(data, offset)
            self._update_crc()
        except Exception as e:
            logger.error(f"Error writing to EEPROM, {e}")


class EKF_EEPROM(Validatable_EEPROM, ProbeableDevice):
    """
    A class used to represent an EKF EEPROM device.

    Structure
    ---------
    The EKF_EEPROM content is structured as follows:

    - `Serial number` (4 bytes, starts at pos 8):
        The serial number of the device.
    - `Manufactured at` (2 bytes, starts at pos 12):
        The date the device was manufactured.
    - `Repaired at` (2 bytes, starts at pos 14):
        The date the device was repaired.
    - `Customer serial number` (4 bytes, starts at pos 32):
        The customer serial number of the device.
    - `Customer configuration block offset pointer` (4 bytes, starts at pos 36):
        The offset pointer to the customer configuration block.
    - `String array` (78 bytes, starts at pos 48):
        An array of strings containing the model, manufacturer, and custom board data of the device.
    - `CRC` (2 bytes, starts at pos 126):
        The CRC value of the EEPROM content.
    - `Raw content` (80 bytes, starts at pos 128):
        Free customizable content for other purposes.


    Note
    ----
    - As the CRC value is stored at the end of the OEM data space, just before the customer configuration block,
      the CRC position is manually set and the :meth:`~ekfsm.devices.eeprom.Validatable_EEPROM._update_content` method is
      overridden.
    - The CRC value is computed using the `CRC-16/XMODEM <https://en.wikipedia.org/wiki/Cyclic_redundancy_check>`_ algorithm.
    - Dates are stored in a proprietary format (2 bytes) and must be decoded using the :meth:`~EKF_EEPROM._decode_date` method.


    See Also
    --------
    `crcmod <https://crcmod.sourceforge.net/>`_


    Important
    ---------
    All data read from the EEPROM should be validated using the @validated decorator.
    This decorator ensures that the data is not corrupted by checking the CRC value.


    Raises
    ------
    DataCorruptionError
        If the CRC validation fails.
    """

    _sernum_index_start = 8
    _sernum_index_end = 12

    _date_mft_index_start = 12
    _date_mft_index_end = 14

    _date_rep_index_start = 14
    _date_rep_index_end = 16

    _customer_serial_index_start = 32
    _customer_serial_index_end = 36

    _customer_config_block_offset_pointer_index = 36

    _str_array_start_offset = 48
    _str_array_end_offset = 126

    def __init__(
        self,
        name: str,
        parent: SysTree | None = None,
        children: list[Device] | None = None,
        abort: bool = False,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(name, parent, children, abort, *args, **kwargs)

    def _update_content(self) -> None:
        super()._update_content()

        # EKF EEPROM content is restricted to 128 bytes, so strip the rest!
        self._firmware_content = self._content[:128]

        # The rest is raw content available for other purposes
        self._raw_content: bytes = self._content[128:]

        # Firmware data without CRC needs to be overriden
        self._data: bytes = (
            self._firmware_content[self._crc_length :]
            if self._crc_pos == "start"
            else self._firmware_content[: -self._crc_length :]
        )
        self._str_list = self._get_string_array()

        self._crc_pos_start = len(self._data)
        self._crc_pos_end = self._crc_pos_start + self._crc_length

    @validated
    def serial(self) -> str:
        """
        Get the serial number of the device to which the EEPROM is attached (the root device).

        Returns
        -------
            The serial number of the root device.
        """
        area = self._content[self._sernum_index_start : self._sernum_index_end]
        sernum = int.from_bytes(area, byteorder="little")
        return str(sernum)

    def write_serial(self, serial: int) -> None:
        """
        Write serial number of the root device to EEPROM.

        Parameters
        ----------
        serial
            The serial number to write to the EEPROM.
        """
        # Check serial number is within bounds
        unsigned_upper_bound = (2**32) - 1
        if serial < 0 or serial > unsigned_upper_bound:
            raise ValueError(
                f"Serial number must be between 0 and {unsigned_upper_bound}"
            )
        serial_bytes = serial.to_bytes(4, byteorder="little")
        self.write(serial_bytes, self._sernum_index_start)

    @validated
    def custom_serial(self) -> str:
        """
        Get the customer serial number of the device to which the EEPROM is attached (the root device).

        Attention
        ---------
        This is a custom - non-OEM - serial number that can be set by the user.

        Returns
        -------
            The customer serial number of the root device.
        """
        area = self._content[
            self._customer_serial_index_start : self._customer_serial_index_end
        ]
        sernum = int.from_bytes(area, byteorder="little")
        return str(sernum)

    def write_custom_serial(self, serial: int) -> None:
        """
        Write customer serial number of the root device to EEPROM.

        Parameters
        ----------
        serial
            The customer serial number to write to the EEPROM.


        Raises
        ------
        ValueError
            If the serial number is not within the bounds of a 32-bit unsigned integer.


        Note
        ----
        Due to space restrictions on storage, the serial number must be a 32-bit unsigned integer.
        """
        # Check serial number is within bounds
        unsigned_upper_bound = (2**32) - 1
        if serial < 0 or serial > unsigned_upper_bound:
            raise ValueError(
                f"Serial number must be between 0 and {unsigned_upper_bound}"
            )

        serial_bytes = serial.to_bytes(4, byteorder="little")
        logger.debug(f"Writing customer serial {serial}")
        self.write(serial_bytes, self._customer_serial_index_start)

    @validated
    def manufactured_at(self) -> date:
        """
        Get the date the device was manufactured.

        Returns
        -------
            The date the device was manufactured.
        """
        area = self._content[self._date_mft_index_start : self._date_mft_index_end]
        encoded_mft_date = area[::-1]
        return self._decode_date(encoded_mft_date)

    @validated
    def repaired_at(self) -> date:
        """
        Get the date the device was repaired.

        Returns
        -------
            The most recent date the device was repaired.
        """
        area = self._content[self._date_rep_index_start : self._date_rep_index_end]
        encoded_rep_date = area[::-1]
        return self._decode_date(encoded_rep_date)

    @validated
    def write_repaired_at(self, date: date) -> None:
        """
        Write the date the device was repaired to EEPROM.

        Parameters
        ----------
        date
            The date the device was repaired.

        Note
        ----
        The date year must be within the range of 1980-2079.

        Attention
        ---------
        The date is stored in a proprietary 2-byte format.

        Raises
        ------
        ValueError
            If the year is not within the range of 1980-2079.
        """
        if date.year < 1980 or date.year > 2079:
            raise ValueError("Year must be within the range of 1980-2079")
        rep_date_bytes = self._encode_date(date)
        logger.debug(f"Writing repair date {date} to EEPROM")
        self.write(rep_date_bytes, self._date_rep_index_start)

    @validated
    def model(self) -> str | None:
        """
        Get the model name of the device to which the EEPROM is attached to (the root device).

        Returns
        -------
            The model name of the device.
        """
        return self._str_list[0] if len(self._str_list) > 0 else None

    @validated
    def vendor(self) -> str | None:
        """
        Get the vendor/manufacturer of the device to which the EEPROM is attached to (the root device).

        Returns
        -------
            The name of the vendor/manufacturer of the device.
        """
        return self._str_list[1] if len(self._str_list) > 1 else None

    @validated
    def custom_board_data(self) -> str | None:
        """
        Get the custom board data of the device.

        Note
        ----
        This is a custom field that can be set by the user.


        Attention
        ---------
        This field is optional and may not be present in the EEPROM content.


        Returns
        -------
            The custom board data of the device as a string, or `None` if the field is not present.
        """
        return None if len(self._str_list) < 3 else self._str_list[2]

    def write_custom_board_data(self, data: str) -> None:
        """
        Write custom board data to EEPROM.

        Important
        ---------
        Due to size limitations, the custom board data should only contain expressive,
        short content like serials, variants or specific codes.

        Parameters
        ----------
        data
            The custom board data to write to the EEPROM.

        Attention
        ---------
        The model and vendor fields are mandatory and must be set before writing custom board data.

        Raises
        ------
        ValueError
            If the model and vendor fields are not set before writing custom board data.
        """
        data_bytes = data.encode("utf-8")
        data_offset = 0
        for s in self._str_list[:2]:
            if s is None:
                raise ValueError(
                    "Model and vendor fields must be set before writing custom board data"
                )
            if isinstance(s, str):
                data_offset += len(s) + 1
        logger.info(f"Writing custom board data {data} to EEPROM")
        self.write(data_bytes, self._str_array_start_offset + data_offset)

    def custom_raw_data(self) -> bytes:
        """
        Get the raw content area data stored in the EEPROM.

        Returns
        -------
            The data contained in the raw content block of the EEPROM.


        Note
        ----
        This area is free for custom data storage and is not included during crc calculations and validations.


        Important
        ---------
        If custom raw data should be stored on EEPROM and
        if it should be protected against corruption, it has to be validated manually.
        """
        return self._raw_content

    def write_custom_raw_data(self, data: bytes) -> None:
        """
        Write custom data to the raw content area of the EEPROM.

        Parameters
        ----------
        data
            The data to write to the raw content area of the EEPROM.
        """
        logger.info(f"Writing {len(data)} bytes to raw content area of EEPROM")
        self.write(data, 128)

    def _get_string_array(self) -> list[str | None]:
        str_array = self._content[
            self._str_array_start_offset : self._str_array_end_offset
        ].split(b"\x00")
        try:
            return [s.decode("utf-8") for s in str_array if s]
        except UnicodeDecodeError:
            return [None, None, None]

    @classmethod
    def _decode_date(cls, encoded_date: Sequence[int]) -> date:
        """
        Decode a date from a proprietary 2-byte format.

        Parameters
        ----------
        encoded_date
            The date to decode.

        Raises
        ------
        ValueError
            If the date is invalid (e.g., 30th Feb).

        Returns
        -------
        date
            The decoded date.
        """
        bdate = int.from_bytes(encoded_date, byteorder="little")

        # Extract the day (bit 0-4)
        day = bdate & 0x1F  # 0x1F is 00011111 in binary (5 bits)

        # Extract the month (bit 5-8)
        month = (bdate >> 5) & 0x0F  # Shift right by 5 and mask with 0x0F (4 bits)

        # Extract the year since 1980 (bit 9-15)
        year = (bdate >> 9) & 0x7F  # Shift right by 9 and mask with 0x7F (7 bits)
        year += 1980  # Add base year (1980)

        # Return a datetime object with the extracted year, month, and day
        try:
            decoded_date = date(year, month, day)
            return decoded_date
        except ValueError:
            raise ValueError(
                f"Invalid date: {day}/{month}/{year}"
            )  # Handle invalid dates, e.g., 30th Feb

    @classmethod
    def _encode_date(cls, date: date) -> bytes:
        """
        Encode a date into a proprietary 2-byte format.

        Parameters
        ----------
        date
            The date to encode.

        Returns
        -------
        bytes
            The encoded date.
        """
        year = date.year - 1980
        month = date.month
        day = date.day

        encoded_date = year << 9 | month << 5 | day
        return encoded_date.to_bytes(2, byteorder="little")

    def _compute_crc(self) -> int:
        return get_crc16_xmodem(self._data)

    def probe(self, *args, **kwargs):
        return self.hw_module.id == self.model()


class EKF_CCU_EEPROM(EKF_EEPROM):
    """
    EKF CCU EEPROM - uses the second part of the EEPROM for chassis inventory and customer area
    """

    _cvendor_index_start = 128
    _cvendor_length = 24

    _cmodel_index_start = 152
    _cmodel_length = 24

    _crevision_index_start = 176
    _crevision_length = 9

    _cserial_index_start = 185
    _cserial_length = 4

    _unit_index_start = 189
    _unit_length = 1

    _customer_area_start = 190
    _customer_area_length = 63

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

    def _update_content(self) -> None:
        super()._update_content()

        # CCU content is the raw content area of the EEPROM
        self._ccu_content: bytes = self._raw_content

        # CCU Firmware data without CRC needs to be overriden
        self._cdata: bytes = (
            self._ccu_content[self._crc_length :]
            if self._crc_pos == "start"
            else self._ccu_content[: -self._crc_length :]
        )
        self._ccrc_pos_start = len(self._cdata) + 128
        self._ccrc_pos_end = self._ccrc_pos_start + self._crc_length

    def _update_ccrc(self) -> None:
        """
        Update the CRC value of the EEPROM content.
        """
        self.ccrc = self._compute_ccrc()

    def _get_ccrc_value(self) -> int:
        return int.from_bytes(
            self._content[self._ccrc_pos_start : self._ccrc_pos_end], byteorder="little"
        )

    @property
    def ccrc(self) -> int:
        """
        Gets or sets the CRC value of the EEPROM content.

        Parameters
        ----------
        value: optional
            The CRC value to write to the EEPROM.

        Returns
        -------
            int
                The CRC value currently stored in the EEPROM if used as *getter*.
            None
                If used as *setter*.


        Caution
        -------
        The *setter* actually writes the CRC value to the EEPROM.


        Warning
        -------
        The *setter* method should be used with caution as it can lead to data corruption if the CRC value is not correct!


        Note
        ----
        The *setter* is usually triggered automatically after a successful write operation
        and in most cases, there is no need to call it manually.
        """
        return self._get_ccrc_value()

    @ccrc.setter
    def ccrc(self, value: int) -> None:
        ccrc_bytes = value.to_bytes(self._crc_length, byteorder="little")
        logger.debug(f"Writing chassis CRC value {value}")
        try:
            super(Validatable_EEPROM, self).write(ccrc_bytes, self._ccrc_pos_start)
        except Exception as e:
            logger.error(f"Error writing CRC value, error: {e}")

    @property
    def valid(self) -> bool:
        """
        Checks if the EEPROM content is valid by comparing the stored CRC value with the computed CRC value.

        Returns
        -------
        bool
            `True` if the EEPROM content is valid, `False` otherwise.
        """
        return self._compute_ccrc() == self.ccrc and super().valid

    @validated
    def cvendor(self) -> str:
        """
        Get the chassis vendor.

        Returns
        -------
            The vendor of the chassis.
        """

        return (
            self._content[
                self._cvendor_index_start : self._cvendor_index_start
                + self._cvendor_length
            ]
            .strip(b"\x00")
            .decode("utf-8")
        )

    def write_cvendor(self, vendor: str) -> None:
        """
        Write the vendor of the chassis to EEPROM.

        Parameters
        ----------
        vendor
            The vendor of the chassis.
        """
        vendor_bytes = vendor.encode("utf-8")
        vendor_fill = b"\x00" * (self._cvendor_length - len(vendor_bytes))
        vendor_bytes += vendor_fill
        logger.info(f"Writing vendor {vendor}")
        self.write(vendor_bytes, self._cvendor_index_start)

    @validated
    def cmodel(self) -> str:
        """
        Get the chassis model.

        Returns
        -------
            The model of the chassis.
        """
        return (
            self._content[
                self._cmodel_index_start : self._cmodel_index_start
                + self._cmodel_length
            ]
            .strip(b"\x00")
            .decode("utf-8")
        )

    def write_cmodel(self, model: str) -> None:
        """
        Write the model of the chassis to EEPROM.

        Parameters
        ----------
        model
            The model of the chassis.
        """
        model_bytes = model.encode("utf-8")
        model_fill = b"\x00" * (self._cmodel_length - len(model_bytes))
        model_bytes += model_fill
        logger.info(f"Writing model {model}")
        self.write(model_bytes, self._cmodel_index_start)

    @validated
    def cserial(self) -> int:
        """
        Get the chassis serial number.

        Returns
        -------
            The serial number of the chassis.
        """
        area = self._content[
            self._cserial_index_start : self._cserial_index_start + self._cserial_length
        ]
        cserial = int.from_bytes(area, byteorder="little")
        return cserial

    def write_cserial(self, serial: int) -> None:
        """
        Write the serial number of the chassis to EEPROM.

        Parameters
        ----------
        serial
            The serial number of the chassis.
        """
        # Check serial number is within bounds
        unsigned_upper_bound = (2**32) - 1
        if serial < 0 or serial > unsigned_upper_bound:
            raise ValueError(
                f"Serial number must be between 0 and {unsigned_upper_bound}"
            )
        serial_bytes = serial.to_bytes(4, byteorder="little")
        logger.info(f"Writing chassis serial {serial}")
        self.write(serial_bytes, self._cserial_index_start)

    @validated
    def crevision(self) -> str:
        """
        Get the revision of the chassis.

        Returns
        -------
            The revision of the chassis.
        """
        return (
            self._content[
                self._crevision_index_start : self._crevision_index_start
                + self._crevision_length
            ]
            .strip(b"\x00")
            .decode("utf-8")
        )

    def write_crevision(self, revision: str) -> None:
        """
        Write the chassis revision.

        Parameters
        ----------
        revision
            The revision of the chassis.
        """
        revision_bytes = revision.encode("utf-8")
        revision_fill = b"\x00" * (self._crevision_length - len(revision_bytes))
        revision_bytes += revision_fill
        logger.info(f"Writing chassis revision {revision}")
        self.write(revision_bytes, self._crevision_index_start)

    @validated
    def unit(self) -> int:
        """
        Get the subsystem unit number.

        Returns
        -------
            The unit number of the subsystem.
        """
        area = self._content[self._unit_index_start]
        unit = int.from_bytes([area], byteorder="little")
        return unit

    @validated
    def version(self) -> int:
        """
        Get the version of the EEPROM data scheme.

        Note
        ----
        If undefined, the version is set to 255 and then defaults to 0.

        Returns
        -------
            The version of the EEPROM data scheme.
        """
        version = self._content[self._ccrc_pos_start - 1]
        return version

    def write_version(self, version: int) -> None:
        """
        Write the version of the EEPROM data scheme.

        Parameters
        ----------
        version
            The version of the EEPROM data scheme.
        """
        if version < 0 or version > 255:
            raise ValueError("Version must be between 0 and 255")

        if version == 255:
            logger.warning("Version 255 is undefined, setting to 0")
            version = 0

        version_bytes = version.to_bytes(1, byteorder="little")
        logger.info(f"Writing version {version}")
        self.write(version_bytes, self._ccrc_pos_start - 1)

    def write_unit(self, unit: int) -> None:
        """
        Write the subsystem unit number.

        Parameters
        ----------
        unit
            The unit number of the subsystem.
        """
        unit_bytes = unit.to_bytes(1, byteorder="little")
        logger.info(f"Writing unit {unit}")
        self.write(unit_bytes, self._unit_index_start)

    @validated
    def customer_area(self) -> bytes:
        """
        Get the customer area of the CCU EEPROM.

        Returns
        -------
            The customer area of the CCU EEPROM.
        """
        return self._content[
            self._customer_area_start : self._customer_area_start
            + self._customer_area_length
        ]

    def write_customer_area(self, data: bytes) -> None:
        """
        Write data to CCU EEPROM customer area.

        Parameters
        ----------
        data
            The data to write to the customer area of the CCU EEPROM.

        Raises
        ------
        ValueError
            If the data exceeds the customer area length.

        Example
        -------
        >>> eeprom = EKF_CCU_EEPROM()
        >>> eeprom.write_customer_area(b"Hello, World!")
        >>> eeprom.customer_area()
        b'Hello, World!'
        """
        if len(data) > self._customer_area_length:
            raise ValueError("Data exceeds customer area length")
        self.write(data, self._customer_area_start)

    def _compute_ccrc(self) -> int:
        return get_crc16_xmodem(self._cdata)

    def write(self, data: bytes, offset: int = 0) -> None:
        try:
            super(Validatable_EEPROM, self).write(data, offset)
        except Exception as e:
            logger.error(f"Error writing to EEPROM, {e}")
        self._update_ccrc()

    def custom_raw_data(self) -> bytes:
        raise NotImplementedError("CCU EEPROM does not have a raw content area")

    def write_custom_raw_data(self, data: bytes) -> None:
        raise NotImplementedError("CCU EEPROM does not have a raw content area")



# Test with config file that doesn't match the sysfs entries
import logging
import pprint
from pathlib import Path

import pytest

from ekfsm.simctrl import enable_simulation, register_gpio_simulations
from ekfsm.system import HWModule, System


@pytest.fixture
def simulation():
    enable_simulation(Path(__file__).parent / "sim" / "sys")
    register_gpio_simulations()


def test_system(simulation):
    config = Path(__file__).parent / "sim" / "test_system_inconsistent.yaml"
    system = System(config)
    pprint.pprint(f"System slots {system.slots}")

    psu = system.psu
    assert psu is not None
    assert isinstance(psu, HWModule)
    assert psu.probe()
    assert psu.inventory.vendor() == "Hitron"
    assert psu.inventory.model() == "HDRC300S-110J-D120E(N)"
    assert psu.main.voltage() == 12.2
    assert psu.sby.current() == 0.5
    slot_info = system.slots.PSU_SLOT.info()
    assert slot_info["is_populated"] is True
    assert slot_info["is_correctly_populated"] is False  # as no desired HWModule type in config

    cpu = system["CPU"]
    assert cpu is not None
    assert isinstance(cpu, HWModule)
    assert cpu.inventory.vendor() == "EKF Elektronik"
    assert cpu.inventory.model() == "SC9-TOCCATA"
    assert cpu.inventory.serial() == "53082029"

    cpu_slot_info = system.slots.SYSTEM_SLOT.info()
    assert cpu_slot_info["name"] == "SYSTEM_SLOT"
    assert cpu_slot_info["is_populated"] is True
    assert cpu_slot_info["is_correctly_populated"] is True

    # sysfs entries are for SRF-FAN, but config file says SUR-LED, so SUR specific devices are not found
    sur = system.slots["SLOT1"].hwmodule
    assert sur is not None
    assert isinstance(sur, HWModule)
    assert sur.board_type == "EKF SRF-FAN"
    assert sur.inventory.vendor() == "EKF Elektronik"
    assert sur.inventory.model() == "SRF-FAN"

    sur_slot = system.slots.SLOT1
    slot_info = sur_slot.info()
    assert slot_info["name"] == "SLOT1"
    assert slot_info["is_populated"] is True
    assert slot_info["is_correctly_populated"] is False
    assert slot_info["actual_hwmodule_type"] == "EKF SRF-FAN"

    slot_info = system.slots.SLOT2.info()
    assert slot_info["name"] == "SLOT2"
    assert slot_info["is_populated"] is False
    assert slot_info["is_correctly_populated"] is False


if __name__ == "__main__":
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )
    # pytest.main([__file__,"--capture=tee-sys"])
    pytest.main([__file__])



from __future__ import annotations

import glob
import os
from pathlib import Path

import ekfsm.core
from ekfsm.core.sysfs import sysfs_root
from ekfsm.devices.generic import Device

# Path to the root of the HWMON sysfs filesystem
HWMON_ROOT = sysfs_root() / Path("class/hwmon")


def find_core_temp_dir(hwmon_dir) -> Path:
    """
    Find the directory containing the coretemp hwmon device.

    Args:
        hwmon_dir: Path to the hwmon directory

    Returns:
        Path to the directory containing the coretemp hwmon device

    Raises:
        FileNotFoundError: If no coretemp directory is found
    """
    # List all 'name' files in each subdirectory of hwmon_dir
    name_files = glob.glob(os.path.join(hwmon_dir, "*", "name"))

    # Search for the file containing "coretemp"
    for name_file in name_files:
        with open(name_file, "r") as file:
            if file.readline().strip() == "coretemp":
                # Return the directory containing this file
                return Path(os.path.dirname(name_file))

    raise FileNotFoundError("No coretemp directory found")


class CoreTemp(Device):
    """
    This class provides an interface to read the CPU core temperature from the HWMON device.

    Note:
    Currently, only the average temperature over all cores is read from the HWMON device.
    """

    def __init__(
        self,
        name: str,
        parent: Device,
        children: list["Device"] | None = None,
        abort: bool = False,
        *args,
        **kwargs,
    ):
        dir = find_core_temp_dir(sysfs_root() / Path("class/hwmon"))
        self.sysfs_device = ekfsm.core.sysfs.SysfsDevice(dir, False)

        super().__init__(name, parent, None, abort, *args, **kwargs)

    def cputemp(self):
        """
        Get the CPU temperature from the HWMON device.

        Returns
        -------
        int
            The CPU temperature in degrees Celsius.
        """
        return self.sysfs.read_int("temp1_input") / 1000



import logging
from pathlib import Path

from ekfsm.system import HWModule, System

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")


def print_inventory(name, inventory):
    print(f"{name}: {inventory.vendor()} {inventory.model()} {inventory.serial()}")


def probe_and_print_inventory(board: HWModule):
    board.probe()
    print_inventory(board.name, board.inventory)


config = Path(__file__).parent / "zip2-comp2.yaml"
system = System(config, abort=False)
print(system)

_psu = system.psu
probe_and_print_inventory(_psu)

_cpu = system.cpu
probe_and_print_inventory(_cpu)

_lan = system.lan
probe_and_print_inventory(_lan)

_ssd1 = system.ssd1
probe_and_print_inventory(_ssd1)

_ser = system.ser
probe_and_print_inventory(_ser)

_ssd2 = system.ssd2
probe_and_print_inventory(_ssd2)

_ssd3 = system.ssd3
probe_and_print_inventory(_ssd3)

_info = system.info
probe_and_print_inventory(_info)

# _ccu = system.ccu
# probe_and_print_inventory(_ccu)


# SUR LED Test
_ser.led_a.set(0, "red")
_ser.led_a.set(1, "yellow")
_ser.led_c.set(0, "white")
_ser.led_d.set(1, "blue")

print(f"SUR TEMP: {_ser.th.temperature()}")
print(f"SUR HUMIDITY: {_ser.th.humidity()}")


# # CCU Test
# # chassis inventory via ccu object
# print_inventory("chassis", _ccu.chassis_inventory)
# # alternative way to get chassis inventory
# print_inventory("chassis", system.inventory)

# print(f"CCU firmware: {_ccu.management.identify_firmware()}")



from typing import Any, List, Tuple

import munch
import yamale  # type: ignore
import yaml

from ekfsm.exceptions import ConfigError

schema_str = """
system_config:
  name: str()
  aggregates: map(required=False)
  slots: list(include('slot'))
---
slot:
  name: str()
  slot_type: regex('^CPCI_S0_(PER|SYS|PSU|UTILITY)$')
  desired_hwmodule_type: str()
  desired_hwmodule_name: str()
  attributes: map(required=False)
"""


def _validate_config(config_file: str) -> None:
    schema = yamale.make_schema(content=schema_str)
    data = yamale.make_data(config_file)
    try:
        yamale.validate(schema, data)
    except yamale.YamaleError:
        raise ConfigError("Error in configuration file")


def _parse_config(config_file: str) -> Any | munch.Munch | List | Tuple:
    try:
        with open(config_file) as file:
            config = yaml.safe_load(file)
            munchified_config = munch.munchify(config)
    except OSError:
        raise ConfigError("Failed to open configuration file: {config_file}")
    return munchified_config


def load_config(config_file: str) -> Any:
    _validate_config(config_file)
    config = _parse_config(config_file)
    return config



from ekfsm.devices.generic import Device
from ekfsm.devices.io4edge import IO4Edge
from io4edge_client.pixelDisplay import Client
from PIL import Image


class PixelDisplay(Device):
    """
    Device class for handling a pixel display.
    """

    def __init__(
        self,
        name: str,
        parent: IO4Edge,
        children: list[Device] | None = None,
        abort: bool = False,
        service_suffix: str | None = None,
        *args,
        **kwargs,
    ):

        super().__init__(name, parent, children, abort, *args, **kwargs)

        self.name = name

        if service_suffix is not None:
            self.service_suffix = service_suffix
        else:
            self.service_suffix = name

        self.service_addr = f"{parent.deviceId}-{self.service_suffix}"

        self.client = Client(self.service_addr)

    def describe(self) -> dict:
        """
        Returns a description of the pixel display.
        """
        describe = self.client.describe()
        desc = {
            "height": describe.height_pixel,
            "width": describe.width_pixel,
            "max_num_of_pixel": describe.max_num_of_pixel,
        }
        return desc

    @property
    def height(self) -> int:
        """
        Returns the height of the pixel display in pixels.
        """
        return self.describe()["height"]

    @property
    def width(self) -> int:
        """
        Returns the width of the pixel display in pixels.
        """
        return self.describe()["width"]

    def off(self) -> None:
        """
        Turn off the pixel display.
        @raises RuntimeError: if the command fails
        @raises TimeoutError: if the command times out
        """
        self.client.set_display_off()

    def display_image(self, path: str) -> None:
        """
        Display an image on the pixel display.
        @raises RuntimeError: if the command fails
        @raises TimeoutError: if the command times out
        """
        with Image.open(path) as img:
            img = img.convert("RGB")
            pix = img.load()

        for i in range(0, 320, 16):
            pix_area = []
            for k in range(0, 16):
                for j in range(0, 240):
                    pix_area.append(pix[j, i + k])
            self.client.set_pixel_area(0, i, 239, pix_area)

    def __repr__(self):
        return f"{self.name}; Service Address: {self.service_addr}"



import pprint
from pathlib import Path
import threading
import time

from ekfsm.system import System

config = Path(__file__).parent / "sq3-only.yaml"
system = System(config, abort=True)

pprint.pprint(f"System slots {system.slots}")

system.print()

cpu = system["CPU"]
cpuB = system.cpu
cpuC = system[0]

assert cpu == cpuB == cpuC

cpu_slot = system.slots["SYSTEM_SLOT"]
cpu_slotB = system.slots.SYSTEM_SLOT
cpu_slotC = system.slots[0]

assert cpu_slot == cpu_slotB == cpu_slotC

cpu.print()
print(f"probing CPU: {cpu.probe()}")
print(
    f"inventory: {cpu.inventory.vendor()} {cpu.inventory.model()} {cpu.inventory.serial()}"
)

sq3 = system["info"]
i4e = sq3.bmc

fw_title, fw_name = i4e.identify_firmware()

print(f"Firmware: {fw_title} {fw_name}")

pixel = i4e.display

pixel.display_image(str(Path(__file__).parent / "sim/SQ3.png"))
# pixel.off()

button_array = i4e.buttons

up = button_array.up
down = button_array.down

up.handler = lambda: print("Up pressed")
down.handler = lambda: print("Down pressed")

stop_event = threading.Event()
button_thread = threading.Thread(target=button_array.read, args=(stop_event,))
button_thread.start()

# ... your main code ...
for i in range(30):
    print(f"Main loop {i}")
    time.sleep(1)

# To stop the thread:
stop_event.set()
button_thread.join()

# # currently only works with busy waiting (timeout=0) or single context
# with up.read() as up_stream:
#     for up_data in up_stream:
#         if up_data:
#             print("Up button pressed")



#! /usr/bin/env python3
# import os
# import sys
# pyright: reportOptionalMemberAccess = false
import logging
from pathlib import Path

import click

from ekfsm.log import ekfsm_logger
from ekfsm.simctrl import enable_simulation, register_gpio_simulations
from ekfsm.system import System

logging.basicConfig(level=logging.WARNING, format="%(levelname)s: %(message)s")
logger = ekfsm_logger(__name__)

sm: System | None = None

__all__ = ("cli", "main", "write", "show")


@click.group()
@click.option("--verbose", "-v", is_flag=True, help="Enable verbose output")
@click.option("--debug", "-d", is_flag=True, help="Enable debug output")
@click.option(
    "--sysfs",
    "-s",
    help="Use custom sysfs dir for simulation mode",
    is_flag=False,
    flag_value="tests/sim/sys",
    type=click.Path(exists=True),
)
@click.option(
    "--config",
    "-c",
    type=click.Path(exists=True),
    required=True,
    help="Path to configuration file",
)
def cli(verbose, debug, sysfs, config):
    global sm
    """POSIX-compliant CLI tool with subcommands
    for the EKF System Management (EKFSM) library.

    This tool provides a command-line interface for managing and
    interacting with the EKF System Management library. It allows
    users to perform various operations related to system management,
    including reading and writing data to the system, and displaying
    information about the system.

    Parameters
    ----------
    verbose
        Enable verbose output. If set, the logging level will be set to INFO.
    debug
        Enable debug output. If set, the logging level will be set to DEBUG.
    sysfs
        Use custom sysfs directory for simulation mode. If set, the
        simulation mode will be enabled and the specified directory
        will be used for sysfs operations.
    config
        Path to the configuration file. This file is required for
        initializing the system. The path should point to a valid
        configuration file in YAML format.
    """
    if verbose:
        logging.getLogger().setLevel(logging.INFO)
        logger.info("Verbose output enabled")
    if debug:
        logging.getLogger().setLevel(logging.DEBUG)
        logger.setLevel(logging.DEBUG)
        logger.debug("Debug output enabled")
    if sysfs:
        logger.warning("Simulation mode enabled")
        enable_simulation(Path(sysfs))
        register_gpio_simulations()

    sm = System(config)


@cli.command()
@click.option(
    "--serial",
    "-s",
    is_flag=False,
    prompt=True,
    type=int,
    help="Write chassis serial number",
)
@click.option("--unit", "-u", is_flag=False, prompt=True, type=int, help="Write chassis unit")
@click.option(
    "--vendor",
    "-n",
    is_flag=False,
    help="Write chassis vendor",
    default="EKF Elektronik",
)
@click.option("--revision", "-r", is_flag=False, help="Write chassis revision")
@click.option("--model", "-m", is_flag=False, help="Write chassis model")
@click.option("--custom", "-c", is_flag=False, help="Write chassis custom information")
@click.option("--version", "-v", is_flag=False, help="Write schema version", default=1)
def write(serial, unit, vendor, revision, model, custom, version):
    """Write data to the system"""
    chassis = sm.ccu.chassis_inventory
    eeprom = sm.ccu.mux.ch00.eeprom

    if serial:
        chassis.write_serial(serial)
    if unit:
        chassis.write_unit(unit)
    if vendor:
        chassis.write_vendor(vendor)
    if revision:
        chassis.write_revision(revision)
    if model:
        chassis.write_model(model)
    if custom:
        chassis.write_customer_area(custom)
    if version:
        eeprom.write_version(version)


@cli.command()
def show():
    """Show information about the system"""
    sm.print()


# Attach subgroups to main CLI
cli.add_command(write)
cli.add_command(show)


def main():
    cli(auto_envvar_prefix="EKFSM")


if __name__ == "__main__":
    main()



#!python
import random
import sys
import time

from ekfsm.lock import Locker


def use_shared_resource(myid: str):
    with open("shared_resource.txt", "w") as f:
        f.write(f"{myid}\n")
    time.sleep(random.random() * 1)  # Simulate some work with the shared resource
    # read the file
    with open("shared_resource.txt", "r") as f:
        data = f.read(len(myid))
        if data != myid:
            raise ValueError(f"file doesn't contain my id, but {data}")
        print("data ok")


def main():
    myid = sys.argv[1]
    print(f"myid is {myid}")

    lock = Locker("shared_resource.txt")

    for _ in range(5):
        with lock.lock():
            use_shared_resource(myid)
        time.sleep(0.5)

    sys.exit(0)


main()



from enum import Enum
from typing import Any

from munch import Munch

from ekfsm.core.components import HWModule


class SlotType(Enum):
    """
    Define the types of slots that can be found in a chassis.

    The following slot types are defined:
    - CPCI_S0_UTILITY: CompactPCI Serial Utility Connector
    - CPCI_S0_SYS: CompactPCI Serial System Slot
    - CPCI_S0_PER: CompactPCI Serial Peripheral Slot
    - CPCI_S0_PSU: CompactPCI Serial Power Supply Slot
    """

    CPCI_S0_UTILITY = 1  # CompactPCI Serial Utility Connector
    CPCI_S0_SYS = 2  # CompactPCI Serial System Slot
    CPCI_S0_PER = 3  # CompactPCI Serial Peripheral Slot
    CPCI_S0_PSU = 4  # CompactPCI Serial Power Supply Slot

    @classmethod
    def from_string(cls, name: str) -> "SlotType":
        try:
            return cls[name]
        except KeyError:
            raise ValueError(f"Invalid {cls.__name__}: {name}")

    def to_string(self) -> str:
        return self.name


class Slot:
    """
    A slot represents a physical slot in a chassis.

    Parameters
    ----------
    name
        The name of the slot, e.g. "SlotA" or "CPCI-SYSTEMSLOT"
    slot_type
        The type of the slot, e.g. SlotType.CPCI_S0_SYS
    desired_hwmodule_type
        The desired type of the hardware module that should be in the slot (currently unused)
    desired_hwmodule_name
        The name to be used for the hardware module instance in the slot. (currently unused)
    master
        The master board of the system
    hwmodule
        The hardware module that is in the slot
    number
        The number of the slot
    attributes
        Additional attributes
    """

    def __init__(
        self,
        name: str,
        slot_type: SlotType,
        desired_hwmodule_type: str,
        desired_hwmodule_name: str,
        number: int,
        hwmodule: HWModule | None = None,
        master: HWModule | None = None,
        attributes: Munch | None = None,
    ) -> None:
        self._name = name
        self.slot_type = slot_type
        self._desired_hwmodule_type = desired_hwmodule_type
        self._desired_hwmodule_name = desired_hwmodule_name
        self.number = number
        self.hwmodule = hwmodule
        self.master = master
        self.attributes = attributes

    def info(self) -> dict[str, Any]:
        """
        Returns a dictionary with information about the slot.

        - name (str): The name of the slot
        - slot_type (str): The type of the slot
        - number (int): The number of the slot
        - desired_hwmodule_type (str): The desired type of the hardware module
        - actual_hwmodule_type (str): The actual type of the hardware module
        - desired_hwmodule_name (str): The desired name of the hardware module
        - is_populated (bool): Is the slot populated?
        - is_correctly_populated (bool): Is the slot correctly populated?
        """
        return {
            "name": self._name,
            "slot_type": self.slot_type.to_string(),
            "number": self.number,
            "desired_hwmodule_type": self._desired_hwmodule_type,
            "actual_hwmodule_type": self.hwmodule.board_type if self.hwmodule else None,
            "desired_hwmodule_name": self._desired_hwmodule_name,
            "is_populated": self.is_populated,
            "is_correctly_populated": self.is_correctly_populated,
        }

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(name={self._name}, slot_type={self.slot_type})"
        )

    @property
    def name(self) -> str:
        """
        Return the name of the slot.
        """
        return self._name

    @property
    def is_populated(self) -> bool:
        """
        Return True if the slot is populated, False otherwise.
        """
        return self.hwmodule is not None

    @property
    def is_correctly_populated(self) -> bool:
        """
        Return True if the slot is populated with the desired hardware module type, False otherwise.
        """
        return (
            self.hwmodule is not None
            and self.hwmodule.board_type.lower() == self._desired_hwmodule_type.lower()
        )


class Slots(Munch):
    """
    A collection of slots.

    Slots are stored in a dictionary-like object, where the key is the slot name and the value is the Slot object.
    Slots can be accessed by name, by number or via an attribute access matching the key.

    Example
    -------
    >>> from ekfsm..core.slots import Slot, Slots, SlotType
    >>> slotA = Slot("SlotA", SlotType.CPCI_S0_PER, "Bla", "Blubb", 3)
    >>> slots = Slots((slotA.name, slotA))
    >>> print(slots[name])
    >>> print(slots.slotA) # attribute access, same as slots[name]
    >>> print(slots[3]) # number access, same as slots.slotA
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def __getitem__(self, key: int | str) -> Slot:
        """
        Get a slot by name, number or attribute access.
        """
        if isinstance(key, int):
            return next(slot for slot in self.values() if slot.number == key)

        return super().__getitem__(key)

    def __setitem__(self, key, value):
        """
        Add a Slot object to the collection.

        Raises
        ------
        ValueError
            if:
            - the value is not a Slot object
            - the key does not match the slot name
            - the slot already exists in collection
            - or the slot number is not unique
        """
        if not isinstance(value, Slot):
            raise ValueError("Only Slot instances can be added to a Slots collection.")
        elif key != value.name:
            raise ValueError("Slot name must match key.")
        elif value in self.values():
            raise ValueError("Slot already exists in collection.")
        elif value.number in [slot.number for slot in self.values()]:
            raise ValueError("Slot number must be unique.")

        return super().__setitem__(key, value)

    def add(self, slot: Slot) -> None:
        """
        Add a Slot object to the collection, where the name of the slot object is used as the key.

        Example
        -------
        >>> from ekfsm.core.slots import Slot, Slots, SlotType
        >>> slotA = Slot("SlotA", SlotType.CPCI_S0_PER, "Bla", "Blubb", 3)
        >>> slots = Slots()
        >>> slots.add(slotA) # add slotA to the collection
        >>> print(slots.SlotA) # attribute access, same as slots["SlotA"]
        """
        self[slot.name] = slot



from ekfsm.devices.generic import Device
from ekfsm.devices.ledArray import LEDArray
from io4edge_client.api.colorLED.python.colorLED.v1alpha1.colorLED_pb2 import Color


class ColorLED(Device):
    """
    Device class for handling a color LED.
    """

    def __init__(
        self,
        name: str,
        parent: LEDArray,
        children: list[Device] | None = None,
        abort: bool = False,
        channel_id: int = 0,
        *args,
        **kwargs,
    ):

        super().__init__(name, parent, children, abort, *args, **kwargs)

        self.name = name
        self.channel_id = channel_id

        self.client = parent.client

    def describe(self):
        pass

    def get(self) -> tuple[Color, bool]:
        """
        Get color LED state.

        Returns
        -------
            Current color and blink state.

        Raises
        ------
        RuntimeError
            if the command fails
        TimeoutError
            if the command times out
        """
        return self.client.get(self.channel_id)

    def set(self, color: Color, blink: bool) -> None:
        """
        Set the color of the color LED.

        Parameters
        ----------
        color : Color
            The color to set the LED to.
        blink : bool
            Whether to blink the LED.

        Raises
        ------
        RuntimeError
            if the command fails
        TimeoutError
            if the command times out
        """
        self.client.set(self.channel_id, color, blink)

        def __repr__(self):
            return f"{self.name}; Channel ID: {self.channel_id}"



from typing import Callable
from ekfsm.devices.generic import Device


class Button(Device):
    """
    Device class for handling a single button as part on array.
    """

    def __init__(
        self,
        name: str,
        parent: Device,
        children: list[Device] | None = None,
        abort: bool = False,
        channel_id: int = 0,
        *args,
        **kwargs,
    ):

        super().__init__(name, parent, children, abort, *args, **kwargs)

        self.channel_id = channel_id

        self._handler: Callable | None = None

    @property
    def handler(self):
        """
        Handle button events with a callback function.
        """
        return self._handler

    @handler.setter
    def handler(self, func: Callable | None, *args, **kwargs):
        """
        Handle button events with a callback function.

        Parameters
        ----------
        func : Callable | None
            The function to call on button events. If None, no function is called.
        """
        if callable(func):
            self._handler = func
        else:
            self._handler = None

    def __repr__(self):
        return f"{self.name}; Channel ID: {self.channel_id}"
